#!/bin/sh

# [1] Environment Settings:
# $DNSDIR := BIND/named configuration directory
#	example:
#		/etc/namedb
# $DNSGET := WWW based contents retriving program (it outputs to stdout)
#	usage: program <url>
#	example:
#		/usr/bin/fetch -q -o -
#		/usr/local/bin/wget -q -O -
DNSDIR=/etc/namedb
DNSGET='/usr/local/bin/wget -q -O -'
export DNSDIR DNSGET
# 

# [2] Servername Setting:
# $myname := fqdn of nameserver (do not write the last '.')
#	example:
#		ns.ipc.tokushima-u.ac.jp
#		ns2.ipc3.tokushima-u.ac.jp
myname=ns.ipc.tokushima-u.ac.jp
#

# [3] Required Directories:
# $DNSDIR/edb  : EDB/DNS working directory
# $DNSDIR/zone : directory for zone files of domain
# $DNSDIR/rev  : directory for zone files of in-addr.arpa
# $DNSDIR/back : directory for backup files
#

# [4] Optional Files:
# $DNSDIR/named.conf-options : options for zone definitions
#	Format: OPTION/name-of-zone/options
#	example:
#	OPTION/basic.med.tokushima-u.ac.jp/also-notify { 150.59.190.3; };
#	OPTION/clin.med.tokushima-u.ac.jp/also-notify { 150.59.190.3; };
#	OPTION/72.59.150.in-addr.arpa/also-notify { 150.59.190.3; };
#

# Step1: Making dummy zone files
touch $DNSDIR/edb/dummy.zone $DNSDIR/edb/dummy.rev

# Step2: Making configurations (named.conf) from EDB/DNS
cd $DNSDIR/edb
./edb-dns-conf.sh $myname
cd $DNSDIR
# If this procedure is correctly executed, the files
#	$DNSDIR/edb/named.conf-body : the zone definition part of named.conf
#	$DNSDIR/edb/*.zone : zone files for domain
#	$DNSDIR/edb/*.rev : zone files for in-addr.arpa
# are generated.

# Step3: Copy generated files to standard directory
if [ -s $DNSDIR/edb/named.conf-body ]; then
	cp $DNSDIR/edb/named.conf-body ./named.conf-tunes
fi
cp $DNSDIR/edb/*.zone $DNSDIR/zone/.
cp $DNSDIR/edb/*.rev $DNSDIR/rev/.

# Step4: Remove dummy zone files
rm -f $DNSDIR/zone/dummy.zone $DNSDIR/rev/dummy.rev

# Step5: Send "reload" command to named.
/usr/sbin/ndc reload > /dev/null

exit 0
