/*
	$Id: dns.c,v 1.1 2005/08/10 10:13:02 alex Exp $
	dns.c--- DNS routines
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: dns.c,v 1.1 2005/08/10 10:13:02 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<ctype.h>
#include	<string.h>
#include	<signal.h>
#include	"edb/edb.h"
#include	"edb/pki.h"
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>
#include	<arpa/nameser.h>
#include	<netdb.h>
#include	<errno.h>
#include	<resolv.h>
#include	"dns.h"

extern	int	h_errno;	/* for resolver errors */
extern	int	errno;	/* general system errors */

static int	default_nscount;
static u_int	default_options;
static struct in_addr	default_nsaddr_list[MAXNS];

		int	dns_initialize					__P((void));
		int	dns_store_default				__P((void));
		int	dns_restore_default				__P((void));
static	int	dns_set_nameserver				__P((char *[], FILE *));
		int	dns_find_NS					__P((char *, char *[], char *[], int *, FILE *));
static	int	dns_query_domain_information_soa	__P((char *, struct DNS_DOMAIN_INFO *, FILE *));
static	int	dns_query_domain_information_ns	__P((char *, struct DNS_DOMAIN_INFO *, FILE *));
		int	dns_query_domain_information		__P((char *, char *nameserver, struct DNS_DOMAIN_INFO *, FILE *));
		int	dns_compare_domain_information	__P((struct DNS_DOMAIN_INFO *, struct DNS_DOMAIN_INFO *));
static	int	skipName						__P((u_char *, u_char *, FILE *));
static	int	skipToData					__P((u_char *, u_short *, u_short *, u_int *, u_short *, u_char *));
static	void	returnCodeError					__P((int, FILE *));
static	void	nsError						__P((int, char *, FILE *));


int
dns_initialize()
{

	res_init();
	dns_store_default();
	return TRUE;
}

int
dns_store_default()
{
	int	i;

	default_nscount	= _res.nscount;
	for(i=0; i<default_nscount; ++i) {
		default_nsaddr_list[i] = _res.nsaddr_list[i].sin_addr;
	}
	default_options = _res.options;
	return TRUE;
}

int
dns_restore_default()
{
	int	i;

	_res.nscount = default_nscount;
	for(i=0; i<default_nscount; ++i) {
		_res.nsaddr_list[i].sin_addr = default_nsaddr_list[i];
	}
	_res.options = default_options;
	return TRUE;
}

static
int
dns_set_nameserver(char *ns_list[], FILE *fp)
{
	struct	hostent	*host;
	int	n;
	char	**ns;

	dns_restore_default();
	_res.options |= RES_RECURSE;
	_res.retry = 4;
	for(ns = ns_list, n=0; *ns && n<MAXNS; ++ns) {
		if(!inet_aton(*ns, &_res.nsaddr_list[n].sin_addr)) {
			host = gethostbyname(*ns);
			if(host == NULL) {
				fprintf(fp, "There is no address for %s\n", *ns);
				dns_restore_default();
				return FALSE;
			}
			memcpy(&_res.nsaddr_list[0].sin_addr, (void *)host->h_addr_list[0], (size_t)host->h_length);
		}
		++n;
	}
	_res.nscount = n;

	_res.options &= ~RES_RECURSE;
	_res.retry = 1;
	return TRUE;
}

int
dns_find_NS(char *domain, char *nameservers[], char *nsList[], int *nsNum, FILE *fp)
{
	union {
		HEADER	hdr;
		u_char	buf[NS_PACKETSZ];
	} response;
	int	responseLen;

	u_char	*cp;
	u_char	*eom;
	u_short	class;
	u_short	type;
	u_int	ttl;
	u_short	dlen;

	int	i, count, dup;

	if(nameservers) dns_set_nameserver(nameservers, fp);

	*nsNum = 0;
	if((responseLen = res_query(domain, C_IN, T_NS, (u_char *)&response, sizeof(response)))<0) {
		nsError(h_errno, domain, fp);
		dns_restore_default();
		return 0;
	}

	eom = response.buf + responseLen;
	cp = response.buf + sizeof(HEADER);

	cp += skipName(cp, eom, fp) + QFIXEDSZ;

	count = ntohs(response.hdr.ancount) + ntohs(response.hdr.nscount);
	while( (--count >= 0) && (cp < eom) && (*nsNum < DNS_NS_LIMIT)) {
		cp += skipToData(cp, &type, &class, &ttl, &dlen, eom);
		if(type==T_NS) {
			nsList[*nsNum] = (char *) malloc((size_t)NS_MAXDNAME);
			if(nsList[*nsNum] == NULL) {
				fprintf(fp, "malloc failed\n");
				exit(1);
			}
			if(dn_expand(response.buf, eom, cp, (char *)nsList[*nsNum], NS_MAXDNAME)<0) {
				fprintf(fp, "dn_expand failed\n");
				exit(1);
			}
			for(i=0, dup=0; (i<*nsNum) && !dup; ++i) {
				dup = !strcasecmp(nsList[i], nsList[*nsNum]);
			}
			if(dup) {
				free(nsList[*nsNum]);
			} else {
				++(*nsNum);
			}
		}
		cp += dlen;
	}
	dns_restore_default();
	return 1;
}

static
int
dns_query_domain_information_soa(char *domain, struct DNS_DOMAIN_INFO *di, FILE *fp)
{
	union	{
		HEADER	hdr;
		u_char	buf[NS_PACKETSZ];
	} query, response;
	int	responseLen, queryLen;

	u_char	*cp;
	u_char	*eom;
	u_short	class;
	u_short	type;
	u_int	ttl;
	u_short	dlen;
	int	nameLen;
	int	i;

	queryLen = res_mkquery(ns_o_query, domain, ns_c_in, ns_t_soa, (u_char *)NULL, 0, NULL, (u_char *)&query, (int)sizeof(query));
	errno = 0;
	if((responseLen = res_send((u_char *)&query, queryLen, (u_char *)&response, (int)sizeof(response))) < 0) {
		if(errno == ECONNREFUSED) {
			fprintf(fp, "There is no name server running\n");
		} else {
			fprintf(fp, "There was no response\n");
		}
		return 0;
	}
	eom = response.buf + responseLen;
	cp = response.buf + sizeof(HEADER);
	if(response.hdr.rcode != NOERROR) {
		returnCodeError((int)response.hdr.rcode, fp);
		return 0;
	}
	if(!response.hdr.aa) {
		fprintf(fp, "not authoritative for %s\n", domain);
		return DNS_NOT_AUTHORITATIVE;
	}
	if(ntohs(response.hdr.ancount) != 1) {
		fprintf(fp, "expected 1 answer, got %d\n", ntohs(response.hdr.ancount));
		return 0;
	}
	cp += skipName(cp, eom, fp) + QFIXEDSZ;
	cp += skipToData(cp, &type, &class, &ttl, &dlen, eom);
	if(type != T_SOA) {
		fprintf(fp, "expected answer type %d, got %d\n", T_SOA, type);
		return 0;
	}
	if((nameLen = dn_expand(response.buf, eom, cp, (char *)di->soa.mname, NS_MAXDNAME))<0) {
		fprintf(fp, "dn_expand failed\n");
		return 0;
	}
	cp += nameLen;
	if((nameLen = dn_expand(response.buf, eom, cp, (char *)di->soa.rname, NS_MAXDNAME))<0) {
		fprintf(fp, "dn_expand failed\n");
		return 0;
	}
	cp += nameLen;
	GETLONG(di->soa.serial, cp);
	GETLONG(di->soa.refresh, cp);
	GETLONG(di->soa.retry, cp);
	GETLONG(di->soa.expire, cp);
	GETLONG(di->soa.minimum, cp);
	return 1;
}

static
int
dns_query_domain_information_ns(char *domain, struct DNS_DOMAIN_INFO *di, FILE *fp)
{
	union	{
		HEADER	hdr;
		u_char	buf[NS_PACKETSZ];
	} query, response;
	int	responseLen, queryLen;

	u_char	*cp;
	u_char	*eom;
	u_short	class;
	u_short	type;
	u_int	ttl;
	u_short	dlen;
	int	nameLen;
	int	i, count, dup;

	di->ns_num = 0;
	queryLen = res_mkquery(ns_o_query, domain, ns_c_in, ns_t_ns, (u_char *)NULL, 0, NULL, (u_char *)&query, (int)sizeof(query));
	errno = 0;
	if((responseLen = res_send((u_char *)&query, queryLen, (u_char *)&response, (int)sizeof(response))) < 0) {
		if(errno == ECONNREFUSED) {
			fprintf(fp, "There is no name server running\n");
		} else {
			fprintf(fp, "There was no response\n");
		}
		return 0;
	}
	eom = response.buf + responseLen;
	cp = response.buf + sizeof(HEADER);
	if(response.hdr.rcode != NOERROR) {
		returnCodeError((int)response.hdr.rcode, fp);
		return 0;
	}
	if(!response.hdr.aa) {
		fprintf(fp, "not authoritative for %s\n", domain);
		return DNS_NOT_AUTHORITATIVE;
	}
	cp += skipName(cp, eom, fp) + QFIXEDSZ;

	count = ntohs(response.hdr.ancount) + ntohs(response.hdr.nscount);
	while( (--count >= 0) && (cp < eom) && (di->ns_num < DNS_NS_LIMIT)) {
		cp += skipToData(cp, &type, &class, &ttl, &dlen, eom);
		if(type==T_NS) {
			if(dn_expand(response.buf, eom, cp, (char *)di->ns[di->ns_num], NS_MAXDNAME)<0) {
				fprintf(fp, "dn_expand failed\n");
				return 0;
			}
			for(i=0, dup=0; i<di->ns_num && !dup; ++i) {
				dup = !strcasecmp(di->ns[i], di->ns[di->ns_num]);
			}
			if(!dup) ++di->ns_num;
		}
		cp += dlen;
	}
	return 1;
}

int
dns_query_domain_information(char *domain, char *nameserver, struct DNS_DOMAIN_INFO *di, FILE *fp)
{
	char		*ns_list[2];
	int	status = 1;

	ns_list[0] = nameserver;
	ns_list[1] = NULL;
	dns_set_nameserver(ns_list, fp);

	strcpy(di->domain, domain);
	strcpy(di->server, nameserver);
	status = dns_query_domain_information_soa(domain, di, fp);
	if(status>0) status = dns_query_domain_information_ns(domain, di, fp);
	dns_restore_default();
	return status;
}

int
dns_compare_domain_information(struct DNS_DOMAIN_INFO *di1, struct DNS_DOMAIN_INFO *di2)
{
	int	ret, i1, i2;

	/* SOA */
	if((ret = strcasecmp(di1->soa.mname, di2->soa.mname))!=0) return ret;
	if((ret = strcasecmp(di1->soa.rname, di2->soa.rname))!=0) return ret;
	if((ret = di1->soa.serial - di2->soa.serial)!=0) return ret;
	if((ret = di1->soa.refresh - di2->soa.refresh)!=0) return ret;
	if((ret = di1->soa.retry - di2->soa.retry)!=0) return ret;
	if((ret = di1->soa.expire - di2->soa.expire)!=0) return ret;
	if((ret = di1->soa.minimum - di2->soa.minimum)!=0) return ret;

	/* NS */
	if((ret = di1->ns_num - di2->ns_num)!=0) return ret;
	for(i1=0; i1<di1->ns_num; ++i1) {
		for(i2=0; i2<di2->ns_num; ++i2) {
			if(!strcasecmp(di1->ns[i1], di2->ns[i2])) break;
		}
		if(i2==di2->ns_num) return 1;
	}

	return 0;
}

static
int
skipName(u_char *cp, u_char *eom, FILE *fp)
{
	int	n;

	if((n = dn_skipname(cp, eom)) < 0) {
		if(fp) fprintf(fp, "dn_skipname failed\n");
		exit(0);
	}
	return	n;
}

static
int
skipToData(u_char *cp, u_short *type, u_short *class, u_int *ttl, u_short *dlen, u_char *eom)
{
	u_char	*tmp_cp = cp;

	tmp_cp += skipName(tmp_cp, eom, NULL);
	GETSHORT(*type, tmp_cp);
	GETSHORT(*class, tmp_cp);
	GETLONG(*ttl, tmp_cp);
	GETSHORT(*dlen, tmp_cp);

	return (tmp_cp - cp);
}

static
void
returnCodeError(int rcode, FILE *fp)
{

	switch(rcode) {
	case FORMERR:
		fprintf(fp, "FORMERR response\n");
		break;
	case SERVFAIL:
		fprintf(fp, "SERVFAIL response\n");
		break;
	case NXDOMAIN:
		fprintf(fp, "NXDOMAIN response\n");
		break;
	case NOTIMP:
		fprintf(fp, "NOTIMP response\n");
		break;
	case REFUSED:
		fprintf(fp, "REFUSED response\n");
		break;
	default:
		fprintf(fp, "unexpected response\n");
		break;
	}
}

static
void
nsError(int error, char *domain, FILE *fp)
{

	switch(error) {
	case HOST_NOT_FOUND:
		fprintf(fp, "Unknown Domain: %s\n", domain);
		break;
	case NO_DATA:
		fprintf(fp, "No records for %s\n", domain);
		break;
	case TRY_AGAIN:
		fprintf(fp, "No response for query\n");
		break;
	default:
		fprintf(fp, "Unexpected error\n");
		break;
	}
}


