/*
	$Id: edb_dns_make.c,v 1.4 2006/01/25 02:29:10 alex Exp alex $
	edb_dns_make.c--- whois database to xml converter
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: edb_dns_make.c,v 1.4 2006/01/25 02:29:10 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<ctype.h>
#include	<string.h>
#include	<signal.h>
#include	"edb/edb.h"
#include	"edb/pki.h"
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>
#include	<arpa/nameser.h>
#include	<netdb.h>
#include	<errno.h>
#include	<resolv.h>
#include	"dns.h"

#define	DNS_CHECK

#define	DOMAIN_C_NAME			"inetdomain.name"
#define	DOMAIN_C_UPPER			"inetdomain.upper"
#define	DOMAIN_C_ORGANIZATION	"inetdomain.organization"
#define	DOMAIN_C_NS				"inetdomain.ns"
#define	DOMAIN_C_MX				"inetdomain.mx"
#define	DOMAIN_C_IPNETWORK		"inetdomain.ipnetwork"

#define	NETWORK_C_ADDRESS		"inetnetwork.address"
#define	NETWORK_C_ORGANIZATION	"inetnetwork.organization"
#define	NETWORK_C_NS			"inetnetwork.ns"

#define	HOST_C_NAME				"inethost.name"
#define	HOST_C_DOMAIN			"inethost.domain"
#define	HOST_C_ADDRESS			"inethost.address"
#define	HOST_C_CNAME			"inethost.cname"
#define	HOST_C_MX				"inethost.mx"
#define	HOST_C_EMAIL			"inethost.email"


#define	MESSAGE_DIR	DNSDIR "/etc/messages"
#define	HELP_FILE	MESSAGE_DIR "/dns-help"

#define	SEARCH_SUBSTR	0
#define	SEARCH_EXACT	1
#define	SEARCH_FIRSTSTR	2
#define	SEARCH_LASTSTR	3

#define	FIELDs		26

#define	ATON(c)	((c)-'a')
#define	NTOA(f)	((f)+'a')

#define	MY_DOMAIN	"TOKUSHIMA-U.AC.JP"

struct LABEL {
	char *japanese;
	char *english;
	int type;
#define	NODEF		(-1)
#define	ANY			0
#define	STAFF		1
	int permit;
};

#define	BILABEL(label, a, j)	((j) ? (label)[ATON((a))].japanese : (label)[ATON((a))].english)

#define	DOMAIN_TYPE		"DOM"
#define	NETWORK_TYPE	"NET"
#define	HOST_TYPE		"HOST"
#define	PERSONAL_TYPE	"PERSON"

#define	DOMAIN		0
#define	NETWORK		1
#define	HOST		2
#define	PERSONAL	3

struct SECTION {
	char	*type;
	char	*name;
	char	*jname;
	struct LABEL *label;
	char	*search;
	char	*relation;
#define	PRINT_LONGFORM	0x00000001
#define	PRINT_RELATION	0x00000002
#define	PRINT_MODIFIED	0x00000010
};
typedef struct SECTION *Section;

static char *prog_name;
#define	END_OF_ZONEINFO	"; END-OF-ZONEINFO"
#define	END_OF_REVINFO	"; END-OF-REVINFO"
#define	END_OF_NSCONF	"; END-OF-NSCONF"

struct REV_INFO {
	in_addr_t	addr;
	EdbTupleInfo	host;
};
typedef	struct REV_INFO	*RevInfo;

#define	EDB_DNS_LINK_TO_DOMAIN_FMT		"/DNS/DOMAIN/%s"
#define	EDB_DNS_LINK_TO_NETWORK_FMT	"/DNS/NETWORK/%s"
#define	EDB_DNS_LINK_TO_HOST_FMT		"/DNS/HOST/%s"

static
char *
time_to_serial(char *src)
{
	int year, mon, day, hour, min, sec;
	int matches;
	char	*dst;

	matches = sscanf(src, "%4d/%2d/%2d %2d:%2d:%2d",
		&year, &mon, &day, &hour, &min, &sec);
	if(matches == 6) {
		asprintf(&dst, "%04d%02d%02d%02d", year, mon, day, hour*4+min/15);
	} else {
		dst = strdup("1994010100");
	}
	return dst;
}

static
in_addr_t
cidr_to_address(char *cidr)
{
	char	*address, *s;

	address = STRDUP(cidr);
	s = strchr(address, '/');
	if(s) *s = '\0';
	return ntohl(inet_addr(address));
}

static
int
cidr_to_mask(char *cidr)
{
	char	*s;
	int	mask = 24;

	s = strchr(cidr, '/');
	if(s) {
		++s;
		mask = ATOI(s);
	}
	return mask;
}

static
Link
edb_dns_pgsql_query(EdbContext ec, EdbTableInfo ti, char *condition, char *order)
{
	Link	l, l_eid, l_tuple;
	char	*query, *where, *orderby;
	int	n;

	if(ti==NULL) return NULL;
	if(STRisVALID(condition)) {
		asprintf(&where, " WHERE %s AND %s", EDB_BASE_AVAILABLE, condition);
	} else {
		asprintf(&where, " WHERE %s", EDB_BASE_AVAILABLE);
	}
	if(STRisVALID(order)) {
		asprintf(&orderby, " ORDER BY %s," EDB_BASE_EID, order);
	} else {
		asprintf(&orderby, " ORDER BY " EDB_BASE_EID);
	}
	asprintf(&query,
		"SELECT " EDB_BASE_EID " FROM %s%s%s",
		TABLE_SN(ti), where, orderby);
	FREE(where);
	FREE(orderby);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_tuple = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	tuple = edb_tuple_get(ec, ATOI(l->d), 0);

		if(!tuple) continue;
		LinkAppend(l_tuple, link_atom_new(tuple, NULL));
	}
	link_free(l_eid);
	return l_tuple;
}

static
EdbTupleInfo
edb_dns_lookup_network_by_cidr(EdbContext ec, char *cidr)
{
	EdbTableInfo	ti;
	EdbTupleInfo	network;
	char	*query, *condition;
	Link	l_eid;
	eid_t	eid;
	int	n;

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetNetwork_XN);
	if(ti==NULL) return NULL;
	asprintf(&query, "SELECT %s FROM %s WHERE %s",
		EDB_BASE_EID,
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionTextEQ("@.address", cidr),
			ConditionTerminate));
	FREE(condition);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	if(link_atoms(l_eid)==1) eid = ATOI(l_eid->d);
	link_free(l_eid);
	network = NULL;
	if(EIDisVALID(eid)) {
		EdbTupleInfo	tuple = edb_tuple_get(ec, eid, 0);

		network = tuple;
		if(tuple) {
			Caption	cap = edb_tuple_make_caption(ec, tuple, 0);

			if(strcmp((char *)cap->c_main, cidr)) edb_trace(ec, "ERROR: %s != %s\n", cap->c_main, cidr);
			edb_caption_free(ec, cap);
		}
	}
	return network;
}


static
char *
edb_dns_domain_fqdn(EdbContext ec, EdbTupleInfo domain)
{
	EdbTupleColumn	tc;
	EdbDatum	d;
	char	*name, *fqdn, *uname;
	EdbTupleInfo	upper;

	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.name");
	name = NULL;
	if(tc && (d=tc->datum) && DatumEnglishIsValid(d)) name = STRDUP((char *)DatumEnglish(d));
	if(!STRisVALID(name)) {
		edb_trace(ec, "ERROR: edb_dns_domain_fqdn: can't get name (EID=%d).\n", TupleEID(domain));
		return NULL;
	}
	uname = NULL;
	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.upper");
	if(tc && (d=tc->datum) && DatumEIDisValid(d)) {
		upper = edb_tuple_get(ec, DatumEID(d), 0);
		uname = edb_dns_domain_fqdn(ec, upper);
	}
	if(uname) {
		asprintf(&fqdn, "%s.%s", name, uname);
	} else {
		fqdn = strdup(name);
	}
	return fqdn;
}

static
char *
edb_dns_network_cidr(EdbContext ec, EdbTupleInfo network)
{
	EdbTupleColumn	tc;
	EdbDatum	d;
	char	*cidr;
	EdbTupleInfo	upper;

	tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.address");
	cidr = NULL;
	if(tc && (d=tc->datum) && DatumEnglishIsValid(d)) cidr = STRDUP((char *)DatumEnglish(d));
	if(!STRisVALID(cidr)) {
		edb_trace(ec, "ERROR: edb_dns_domain_fqdn: can't get cidr address (EID=%d).\n", TupleEID(network));
		return NULL;
	}
	return strdup(cidr);
}

static
char *
edb_dns_host_fqdn(EdbContext ec, EdbTupleInfo host)
{
	EdbTupleColumn	tc;
	EdbDatum	d;
	char	*name, *fqdn, *dname;
	EdbTupleInfo	domain;

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.name");
	name = NULL;
	if(tc && (d=tc->datum) && DatumEnglishIsValid(d)) name = STRDUP((char *)DatumEnglish(d));
	if(!STRisVALID(name)) {
		edb_trace(ec, "ERROR: edb_dns_host_fqdn: can't get name (EID=%d).\n", TupleEID(host));
		return NULL;
	}
	dname = NULL;
	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.domain");
	if(tc && (d=tc->datum) && DatumEIDisValid(d)) {
		domain = edb_tuple_get(ec, DatumEID(d), 0);
		dname = edb_dns_domain_fqdn(ec, domain);
	}
	if(dname) {
		asprintf(&fqdn, "%s.%s", name, dname);
	} else {
		fqdn = strdup(name);
	}
	return fqdn;
}

#define	OMIT_MY_DOMAIN	0x00000001

static
void
edb_dns_link_to_domain(EdbContext ec, EdbTupleInfo domain, int mode, Doc doc)
{
	char	*fqdn, *s;

	fqdn = edb_dns_domain_fqdn(ec, domain);
	str_capitalize1(fqdn);
	doc_printf(doc, "<a href=\"" EDB_DNS_LINK_TO_DOMAIN_FMT "\">", fqdn);
	if((mode&OMIT_MY_DOMAIN) && (s=strcasestr(fqdn, MY_DOMAIN))) {
		*s = '\0';
		doc_text_puts(doc, fqdn);
		doc_text_puts(doc, "--");
	} else {
		doc_text_puts(doc, fqdn);
	}
	doc_puts(doc, "</a>");
	FREE(fqdn);
}

static
void
edb_dns_link_to_network(EdbContext ec, EdbTupleInfo network, int mode, Doc doc)
{
	char	*cidr;

	cidr = edb_dns_network_cidr(ec, network);
	str_capitalize1(cidr);
	doc_printf(doc, "<a href=\"" EDB_DNS_LINK_TO_NETWORK_FMT "\">%s</a>", cidr, cidr);
	FREE(cidr);
}

static
void
edb_dns_link_to_host(EdbContext ec, EdbTupleInfo host, int mode, Doc doc)
{
	char	*fqdn, *s;

	fqdn = edb_dns_host_fqdn(ec, host);
	str_minuscule1(fqdn);
	doc_printf(doc, "<a href=\"" EDB_DNS_LINK_TO_HOST_FMT "\">", fqdn);
	if((mode&OMIT_MY_DOMAIN) && (s=strcasestr(fqdn, MY_DOMAIN))) {
		*s = '\0';
		doc_text_puts(doc, fqdn);
		doc_text_puts(doc, "--");
	} else {
		doc_text_puts(doc, fqdn);
	}
	doc_puts(doc, "</a>");
	FREE(fqdn);
}

static
char *
edb_dns_host_address(EdbContext ec, EdbTupleInfo host)
{
	EdbTupleColumn	tc;
	EdbDatum	d;
	char	*address;

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.address");
	if(tc && (d=tc->datum) && DatumEnglishIsValid(d)) return strdup((char *)DatumEnglish(d));
	return NULL;
}

static
EdbTupleInfo
edb_dns_lookup_domain_by_fqdn(EdbContext ec, char *fqdn)
{
	EdbTableInfo	ti;
	EdbTupleInfo	domain;
	char	*query, *condition;
	char	*s, *name;
	Link	l, l_eid;
	int	n;

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetDomain_XN);
	if(ti==NULL) {
		edb_trace(ec, "ERROR: edb_dns_lookup_domain_by_fqdn: can't get %s\n", EDB_TABLE_InetDomain_XN);
		return NULL;
	}
	fqdn = STRDUP(fqdn);
	str_capitalize1(fqdn);
	name = STRDUP(fqdn);
	s = strchr(name, '.');
	if(s) *s = '\0';
	asprintf(&query, "SELECT %s FROM %s WHERE %s",
		EDB_BASE_EID,
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionTextEQ("@.name", name),
			ConditionTerminate));
	FREE(condition);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	domain = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	tuple = edb_tuple_get(ec, ATOI(l->d), 0);
		char	*fqdn2;

		if(!tuple) continue;
		fqdn2 = edb_dns_domain_fqdn(ec, tuple);
		if(STRisVALID(fqdn2)) {
			if(!strcasecmp(fqdn2, fqdn)) {
				if(domain) edb_trace(ec, "ERROR: edb_dns_lookup_domain_by_fqdn: EID is duplicated.\n");
				domain = tuple;
			}
		}
		FREE(fqdn2);
	}
	link_free(l_eid);
	return domain;
}

static
EdbTupleInfo
edb_dns_lookup_host_by_fqdn(EdbContext ec, char *fqdn)
{
	EdbTableInfo	ti;
	EdbTupleInfo	host;
	char	*query, *condition;
	char	*s, *name;
	Link	l, l_eid;
	int	n;

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetHost_XN);
	if(ti==NULL) {
		edb_trace(ec, "ERROR: edb_dns_lookup_host_by_fqdn: can't get %s\n", EDB_TABLE_InetHost_XN);
		return NULL;
	}
	fqdn = STRDUP(fqdn);
	str_capitalize1(fqdn);
	name = STRDUP(fqdn);
	s = strchr(name, '.');
	if(s) *s = '\0';
	asprintf(&query, "SELECT %s FROM %s WHERE %s",
		EDB_BASE_EID,
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionTextEQ("@.name", name),
			ConditionTerminate));
	FREE(condition);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	host = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	tuple = edb_tuple_get(ec, ATOI(l->d), 0);
		char	*fqdn2;

		if(!tuple) continue;
		fqdn2 = edb_dns_host_fqdn(ec, tuple);
		if(STRisVALID(fqdn2)) {
			if(!strcasecmp(fqdn2, fqdn)) {
				if(host) edb_trace(ec, "ERROR: edb_dns_lookup_host_by_fqdn: EID is duplicated.\n");
				host = tuple;
			}
		}
		FREE(fqdn2);
	}
	link_free(l_eid);
	return host;
}

static
Link
edb_dns_lookup_domain_by_name(EdbContext ec, char *name, int find_mode)
{
	EdbTableInfo	ti;
	char	*query, *condition;
	Link	l, l_eid, l_domain;
	int	n, len = strlen(name);

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetDomain_XN);
	if(ti==NULL) {
		edb_trace(ec, "ERROR: edb_dns_lookup_domain_by_fqdn: can't get %s\n", EDB_TABLE_InetDomain_XN);
		return NULL;
	}
	name = STRDUP(name);
	str_capitalize1(name);
	asprintf(&query, "SELECT %s FROM %s WHERE %s",
		EDB_BASE_EID,
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionTerminate));
	FREE(condition);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_domain = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	tuple = edb_tuple_get(ec, ATOI(l->d), 0);
		char	*name2;
		int	len2;

		if(!tuple) continue;
		name2 = edb_dns_domain_fqdn(ec, tuple);
		if(STRisVALID(name2) && (len2=strlen(name2))>=len) {
			switch(find_mode) {
			case SEARCH_SUBSTR:
				if(strcasestr(name2, name)) LinkAppend(l_domain, link_atom_new(tuple, NULL));
				break;
			case SEARCH_EXACT:
				if(!strcasecmp(name2, name)) LinkAppend(l_domain, link_atom_new(tuple, NULL));
				break;
			case SEARCH_FIRSTSTR:
				if(!strncasecmp(name2, name, len)) LinkAppend(l_domain, link_atom_new(tuple, NULL));
				break;
			case SEARCH_LASTSTR:
				if(!strcasecmp(&name2[len2-len], name)) LinkAppend(l_domain, link_atom_new(tuple, NULL));
				break;
			}
		}
		FREE(name2);
	}
	link_free(l_eid);
	return l_domain;
}

static
Link
edb_dns_lookup_host_by_name(EdbContext ec, char *name, int find_mode)
{
	EdbTableInfo	ti;
	Link	l_host;
	char	*query, *condition;
	Link	l, l_eid;
	int	n, len = strlen(name);

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetHost_XN);
	if(ti==NULL) {
		edb_trace(ec, "ERROR: edb_dns_lookup_host_by_name: can't get %s\n", EDB_TABLE_InetHost_XN);
		return NULL;
	}
	name = STRDUP(name);
	str_capitalize1(name);
	asprintf(&query, "SELECT %s FROM %s WHERE %s",
		EDB_BASE_EID,
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionTerminate));
	FREE(condition);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_host = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	tuple = edb_tuple_get(ec, ATOI(l->d), 0);
		char	*name2;
		int	len2;

		if(!tuple) continue;
		name2 = edb_dns_host_fqdn(ec, tuple);
		if(STRisVALID(name2) && (len2=strlen(name2))>=len) {
			switch(find_mode) {
			case SEARCH_SUBSTR:
				if(strcasestr(name2, name)) LinkAppend(l_host, link_atom_new(tuple, NULL));
				break;
			case SEARCH_EXACT:
				if(!strcasecmp(name2, name)) LinkAppend(l_host, link_atom_new(tuple, NULL));
				break;
			case SEARCH_FIRSTSTR:
				if(!strncasecmp(name2, name, len)) LinkAppend(l_host, link_atom_new(tuple, NULL));
				break;
			case SEARCH_LASTSTR:
				if(!strcasecmp(&name2[len2-len], name)) LinkAppend(l_host, link_atom_new(tuple, NULL));
				break;
			}
		}
		FREE(name2);
	}
	link_free(l_eid);
	return l_host;
}

static
Link
edb_dns_lookup_host_by_address(EdbContext ec, char *address, int find_mode)
{
	EdbTableInfo	ti;
	Link	l_host;
	char	*query, *condition;
	Link	l, l_eid;
	int	n, len = strlen(address);

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetHost_XN);
	if(ti==NULL) {
		edb_trace(ec, "ERROR: edb_dns_lookup_host_by_address: can't get %s\n", EDB_TABLE_InetHost_XN);
		return NULL;
	}
	address = STRDUP(address);
	str_capitalize1(address);
	asprintf(&query, "SELECT %s FROM %s WHERE %s",
		EDB_BASE_EID,
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionTerminate));
	FREE(condition);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_host = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	tuple = edb_tuple_get(ec, ATOI(l->d), 0);
		char	*address2;
		EdbTupleColumn	tc;
		int	len2;
		EdbDatum	datum;

		if(!tuple) continue;
		address2 = NULL;
		tc = edb_tuple_seek_column_by_xmlname(ec, tuple, (utf8_t *)"@.address");
		ForTCDatum(datum, tc) {
			if(!DatumEnglishIsValid(datum)) continue;
			address2 = (char *)DatumEnglish(datum);
			if(STRisVALID(address2) && (len2=strlen(address2))>=len) {
				int match = FALSE;
				switch(find_mode) {
				case SEARCH_SUBSTR:
					if(strcasestr(address2, address)) match = TRUE;
					break;
				case SEARCH_EXACT:
					if(!strcasecmp(address2, address)) match = TRUE;
					break;
				case SEARCH_FIRSTSTR:
					if(!strncasecmp(address2, address, len)) match = TRUE;
					break;
				case SEARCH_LASTSTR:
					if(!strcasecmp(&address2[len2-len], address)) match = TRUE;
					break;
				}
				if(match) {
					LinkAppend(l_host, link_atom_new(tuple, NULL));
					break;
				}
			}
		}
	}
	link_free(l_eid);
	return l_host;
}

static
EdbCatalogue
edb_dns_lookup_iphosts_by_EID(EdbContext ec, eid_t eid, int recurse)
{
	EdbTupleInfo	host;
	EdbTupleColumn	tc;
	EdbCatalogue	ca_hosts;
	EdbDatum	datum;

	if(--recurse<=0) return NULL;

	host = edb_tuple_get(ec, eid, 0);
	if(host==NULL) return NULL;

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.address");
	ForTCDatum(datum, tc) {
		if(!DatumEnglishIsValid(datum)) continue;
		ca_hosts = edb_catalogue_new();
		edb_catalogue_append(ca_hosts, TupleEID(host), host);
		return ca_hosts;
	}

	ca_hosts = NULL;
	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.cname");
	ForTCDatum(datum, tc) {
		EdbCatalogue	ca1, ca2;

		if(!DatumEIDisVALID(datum)) continue;
		ca1= edb_dns_lookup_iphosts_by_EID(ec, DatumEID(datum), recurse);
		if(ca1==NULL) {
		} else if(ca_hosts==NULL) {
			ca_hosts = ca1;
		} else {
			ca2 = edb_catalogue_merge(ca_hosts, ca1);
			edb_catalogue_free(ca_hosts);
			edb_catalogue_free(ca1);
			ca_hosts = ca2;
		}
	}
	return ca_hosts;
}

static
EdbCatalogue
edb_dns_lookup_iphosts_by_TC(EdbContext ec, EdbTupleColumn tc)
{
	EdbCatalogue	ca_hosts = edb_catalogue_new();
	EdbDatum	datum;

	ForTCDatum(datum, tc) {
		EdbCatalogue	ca, ca_tmp;

		if(!DatumEIDisVALID(datum)) continue;
		ca = edb_dns_lookup_iphosts_by_EID(ec, DatumEID(datum), ec->max_mapto_recurse);
		if(ca) {
			ca_tmp = edb_catalogue_merge(ca_hosts, ca);
			edb_catalogue_free(ca_hosts);
			edb_catalogue_free(ca);
			ca_hosts = ca_tmp;
		}
	}
	return ca_hosts;
}


static
Link
edb_dns_lookup_domains_by_domain(EdbContext ec, EdbTupleInfo domain)
{
	EdbTableInfo	ti;
	EdbColumnInfo	ci;
	char	*query, *condition;
	char	*s, *name;
	Link	l, l_eid, l_domains;
	eid_t	eid = 0;
	int	n;

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetDomain_XN);
	if(ti==NULL) {
		edb_trace(ec, "ERROR: edb_dns_lookup_domains_by_domain: can't get %s\n", EDB_TABLE_InetDomain_XN);
		return NULL;
	}
	ci = edb_table_seek_column_by_xmlname(ec, ti, (utf8_t *)"@.name");
	asprintf(&query, "SELECT " EDB_BASE_EID " FROM %s WHERE %s ORDER BY %s," EDB_BASE_EID,
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionEIDMatch("@.upper", TupleEID(domain)),
			ConditionTerminate),
		ColumnSN(ci));
	FREE(condition);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_domains = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	tuple = edb_tuple_get(ec, ATOI(l->d), 0);

		if(!tuple) continue;
		LinkAppend(l_domains, link_atom_new(tuple, NULL));
	}
	link_free(l_eid);
	return l_domains;
}

static
Link
edb_dns_lookup_hosts_by_domain(EdbContext ec, EdbTupleInfo domain)
{
	EdbTableInfo	ti;
	EdbColumnInfo	ci;
	char	*query, *condition;
	char	*s, *name;
	Link	l, l_eid, l_hosts;
	eid_t	eid = 0;
	int	n;

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetHost_XN);
	if(ti==NULL) {
		edb_trace(ec, "ERROR: edb_dns_lookup_hosts_by_domain: can't get %s\n", EDB_TABLE_InetHost_XN);
		return NULL;
	}
	ci = edb_table_seek_column_by_xmlname(ec, ti, (utf8_t *)"@.name");
	asprintf(&query, "SELECT " EDB_BASE_EID " FROM %s WHERE %s ORDER BY %s," EDB_BASE_EID,
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionEIDMatch("@.domain", TupleEID(domain)),
			ConditionTerminate),
		ColumnSN(ci));
	FREE(condition);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_hosts = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	tuple = edb_tuple_get(ec, ATOI(l->d), 0);

		if(!tuple) continue;
		LinkAppend(l_hosts, link_atom_new(tuple, NULL));
	}
	link_free(l_eid);
	return l_hosts;
}

static
Link
edb_dns_lookup_hosts_by_network(EdbContext ec, EdbTupleInfo network)
{
	EdbTableInfo	ti;
	char	*query, *condition;
	char	*s, *name;
	Link	l, l_eid, l_hosts;
	eid_t	eid = 0;
	int	n;
	in_addr_t	netaddr;
	EdbTupleColumn	tc;
	EdbDatum	datum;
	char	*cidr, *re;
	int	mask;

	tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.address");
	if(tc==NULL || (datum=tc->datum)==NULL) return NULL;
	cidr = (char *)DatumEnglish(datum);
	if(!STRisVALID(cidr)) return NULL;
	cidr = STRDUP(cidr);
	netaddr	= cidr_to_address(cidr);
	mask	= cidr_to_mask(cidr);
	if(mask==8) {
		asprintf(&re, "^%d[.]", (netaddr>>24)&0xff);
	} else if(mask==16) {
		asprintf(&re, "^%d[.]%d[.]", (netaddr>>24)&0xff, (netaddr>>16)&0xff);
	} else {
		asprintf(&re, "^%d[.]%d[.]%d[.]", (netaddr>>24)&0xff, (netaddr>>16)&0xff, (netaddr>>8)&0xff);
	}

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetHost_XN);
	if(ti==NULL) {
		edb_trace(ec, "ERROR: edb_dns_lookup_hosts_by_network: can't get %s\n", EDB_TABLE_InetHost_XN);
		return NULL;
	}
	asprintf(&query, "SELECT " EDB_BASE_EID " FROM %s WHERE %s ORDER BY " EDB_BASE_EID,
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionRegexTextEQ("@.address", re),
			ConditionTerminate));
	FREE(condition);
	FREE(re);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_hosts = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	tuple = edb_tuple_get(ec, ATOI(l->d), 0);

		if(!tuple) continue;
		LinkAppend(l_hosts, link_atom_new(tuple, NULL));
	}
	link_free(l_eid);
	return l_hosts;
}

static
int
edb_dns_host_to_zoneinfo(EdbContext ec, EdbTupleInfo domain, EdbTupleInfo host, FILE *fp)
{
	Link l, l_address;
	int n;
	char	*name;
	int preference;
	char *s;
	EdbTupleColumn	tc;
	EdbDatum	datum;

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.name");
	if(tc==NULL || tc->datum==NULL) return FALSE;
	datum = tc->datum;
	if(!DatumEnglishIsValid(datum)) return FALSE;
	name = STRDUP((char *)DatumEnglish(datum));
	str_minuscule1(name);

	if((tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.address")) && tc->datum) {
		char	*address, *name1;

		name1 = name;
		ForTCDatum(datum, tc) {
			if(!DatumEnglishIsValid(datum)) continue;
			address = STRDUP((char *)DatumEnglish(datum));
			str_minuscule1(address);
			fprintf(fp, (strlen(name1)>=8) ? "%s\tIN\tA\t%s\n" : "%s\t\tIN\tA\t%s\n", name1, address);
			name1 = "";
		}

		datum = NULL;
		tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.mx");
		if(tc) datum = tc->datum;
		if(datum==NULL || !DatumEIDisVALID(datum)) {
			tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.mx");
			if(tc) datum = tc->datum;
		}
		if(datum && DatumEIDisVALID(datum)) {
			EdbCatalogue	ca_mx = edb_dns_lookup_iphosts_by_EID(ec, DatumEID(datum), ec->max_mapto_recurse);

			if(ca_mx) {
				Link	lm;

				ForCatalogueLink(lm, ca_mx) {
					char	*fqdn;

					host = CatalogueObject(lm);
					fqdn = edb_dns_host_fqdn(ec, host);
					str_minuscule1(fqdn);
					preference = 10;
					fprintf(fp, "\t\tIN\tMX\t%d\t%s.\n", preference, fqdn);
					FREE(fqdn);
					break;
				}
				edb_catalogue_free(ca_mx);
			}
		}
	} else if((tc=edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.cname")) && tc->datum) {
		EdbCatalogue	ca_cname;
		char	*name1, *cname, *address;
		char	*fqdnname, *domainname;

		domainname = edb_dns_domain_fqdn(ec, domain);
		asprintf(&fqdnname, "%s.%s", name, domainname);
		FREE(domainname);
		ca_cname = edb_dns_lookup_iphosts_by_TC(ec, tc);
		name1 = name;
		if(edb_catalogue_size(ca_cname)==1) {
			ForCatalogueLink(l, ca_cname) {
				cname = edb_dns_host_fqdn(ec, CatalogueObject(l));
				str_minuscule1(cname);
				if(!strcasecmp(cname, fqdnname)) continue;
				fprintf(fp, (strlen(name1)>=8) ? "%s\tIN\tCNAME\t%s.\n" : "%s\t\tIN\tCNAME\t%s.\n", name1, cname);
				FREE(cname);
				name1 = "";
			}
		} else {
			ForCatalogueLink(l, ca_cname) {
				address = edb_dns_host_address(ec, CatalogueObject(l));
				str_minuscule1(address);
				fprintf(fp, (strlen(name1)>=8) ? "%s\tIN\tA\t%s\n" : "%s\t\tIN\tA\t%s\n", name1, address);
				FREE(address);
				name1 = "";
			}
		}
		FREE(fqdnname);
		edb_catalogue_free(ca_cname);
	}
	return TRUE;
}

static
int
edb_dns_domain_to_zoneinfo(EdbContext ec, EdbTupleInfo domain, EdbTupleInfo sdomain, EdbCatalogue ca_host_printed, FILE *fp)
{
	Link l;
	char	*name;
	char *s;
	EdbTupleColumn	tc;
	EdbDatum	datum;
	char	*domainname = edb_dns_domain_fqdn(ec, domain);
	int	domainname_len = strlen(domainname);

	tc = edb_tuple_seek_column_by_xmlname(ec, sdomain, (utf8_t *)"@.name");
	if(tc==NULL || tc->datum==NULL) return FALSE;
	datum = tc->datum;
	if(!DatumEnglishIsValid(datum)) return FALSE;
	name = STRDUP((char *)DatumEnglish(datum));
	str_minuscule1(name);

	tc = edb_tuple_seek_column_by_xmlname(ec, sdomain, (utf8_t *)"@.ns");
	if(tc) {
		EdbCatalogue	ca_ns = edb_dns_lookup_iphosts_by_TC(ec, tc);

		ForCatalogueLink(l, ca_ns) {
			EdbTupleInfo	ns = CatalogueObject(l);
			char	*fqdn;

			if(ns==NULL) continue;
			fqdn = edb_dns_host_fqdn(ec, ns);
			if(STRisVALID(fqdn)) {
				str_minuscule1(fqdn);
				fprintf(fp, (strlen(name)>=8) ? "%s\tIN\tNS\t%s.\n" : "%s\t\tIN\tNS\t%s.\n", name, fqdn);
				name = "";
			}
			FREE(fqdn);
		}
		ForCatalogueLink(l, ca_ns) {
			EdbTupleInfo	ns = CatalogueObject(l);
			char	*fqdn;

			if(ns==NULL) continue;
			fqdn = edb_dns_host_fqdn(ec, ns);
			str_minuscule1(fqdn);
			if(!edb_catalogue_lookup(ca_host_printed, TupleEID(ns))) {
				char	*nsname;
				int	nsname_len;
				char *address;

				nsname = STRDUP(fqdn);
				nsname_len = strlen(nsname);
				if(nsname_len>domainname_len
					&& nsname[nsname_len-domainname_len-1]=='.'
					&& !strcasecmp(&nsname[nsname_len-domainname_len], domainname)) {
					nsname[nsname_len-domainname_len-1] = '\0';
					nsname_len = strlen(nsname);
					address = edb_dns_host_address(ec, ns);
					if(address) {
						fprintf(fp, (nsname_len>=8) ? "%s\tIN\tA\t%s\n" : "%s\t\tIN\tA\t%s\n", nsname, address);
						edb_catalogue_append(ca_host_printed, TupleEID(ns), ns);
					}
					FREE(address);
				}
			}
			FREE(fqdn);
		}
		edb_catalogue_free(ca_ns);
	}
	return TRUE;
}

#if 0
/*#hosts_compare_by_name--- hosts compare */
static
int
hosts_compare_by_name(const void *v1, const void *v2)
{
	Wdb	*h1 = (Wdb *)v1;
	Wdb	*h2 = (Wdb *)v2;
	Link ip1, ip2;

	ip1 = (*h1)->field[ATON('b')];
	ip2 = (*h2)->field[ATON('b')];

	if(ip1 && ip2) {
#if 0
		return inet_addr(ip1->d) - inet_addr(ip2->d);
#else
		in_addr_t	v1, v2;

		v1 = inet_addr(ip1->d);
		v2 = inet_addr(ip2->d);
		return ntohl(v1) - ntohl(v2);
#endif
	} else if(ip1) {
		return 1;
	} else if(ip2) {
		return -1;
	} else {
		return 0;
	}
}

/*#hosts_sort_by_name--- sort list */
static
Link
hosts_sort_by_name(Link hosts)
{
	int i, n;
	Wdb *h;
	Link l;

	n = link_atoms(hosts);
	h = (Wdb *)alloca(sizeof(Wdb)*n);
	if( h == NULL ) return NULL;
	for(i=0, l=hosts; i<n; ++i, l=l->n) h[i] = (Wdb)l->d;

	qsort(h, n, sizeof(Wdb), hosts_compare_by_name);

	for(i=0, l=hosts; i<n; ++i, l=l->n) (Wdb)l->d = h[i];

	return hosts;
}
#endif

static
Link
edb_dns_zone_get_hosts(EdbContext ec, EdbTupleInfo domain)
{
	Link	l, l_hosts, l_tuples;

	l_tuples = edb_dns_lookup_hosts_by_domain(ec, domain);
	l_hosts = NULL;
	ForLink(l, l_tuples) {
		EdbTupleInfo	host = l->d;
		EdbTupleColumn	tc;
		EdbDatum	datum;

		if(host==NULL) continue;
		tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.domain");
		ForTCDatum(datum, tc) {
			if(DatumEID(datum)==TupleEID(domain)) {
				LinkAppend(l_hosts, link_atom_new(host, NULL));
				break;
			}
		}
	}
	link_free(l_tuples);
	return l_hosts;
}

static
Link
edb_dns_zone_get_subdomains(EdbContext ec, EdbTupleInfo domain)
{
	Link	l, l_domains, l_tuples;

	l_tuples = edb_dns_lookup_domains_by_domain(ec, domain);
	l_domains = NULL;
	ForLink(l, l_tuples) {
		EdbTupleInfo	sdomain = l->d;
		EdbTupleColumn	tc;
		EdbDatum	datum;

		if(sdomain==NULL) continue;
		tc = edb_tuple_seek_column_by_xmlname(ec, sdomain, (utf8_t *)"@.upper");
		ForTCDatum(datum, tc) {
			if(DatumEID(datum)==TupleEID(domain)) {
				LinkAppend(l_domains, link_atom_new(sdomain, NULL));
				break;
			}
		}
	}
	link_free(l_tuples);
	return l_domains;
}

static
int
edb_dns_make_zoneinfo(EdbContext ec, char *dn, FILE *fp)
{
	Link l, l_hosts, l_domains;
	int preference;
	char *domainname;
	char last_modified[BUFSIZ];
	char *serial, *ns1;
	eid_t	domain_eid;
	EdbTupleInfo	domain, host;
	EdbTupleColumn	tc_ns, tc_mx;
	EdbDatum	d;
	EdbCatalogue	ca_host_printed = edb_catalogue_new();
	EdbCatalogue	ca_ns, ca_mx_printed;
	EdbDatum	datum;

	domain = edb_dns_lookup_domain_by_fqdn(ec, dn);
	if(domain==NULL) {
		fprintf(fp, "; can't get domain information %s\n", dn);
		return FALSE;
	}

	domainname = edb_dns_domain_fqdn(ec, domain);
	str_minuscule1(domainname);

	tc_ns = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.ns");
	tc_mx = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.mx");
	if(tc_ns==NULL || tc_mx==NULL) {
		fprintf(fp, "; No NS or MX is defined in domain %s\n", dn);
		FREE(domainname);
		return FALSE;
	}
	ns1 = NULL;
	if(tc_ns && (d=tc_ns->datum) && DatumEIDisVALID(d)) {
		host = edb_tuple_get(ec, DatumEID(d), 0);
		if(host) ns1 = edb_dns_host_fqdn(ec, host);
	}
	if(!STRisVALID(ns1)) {
		fprintf(fp, "; No NS is defined in domain %s\n", dn);
		FREE(domainname);
		return FALSE;
	}
	str_minuscule1(ns1);
	serial = time_to_serial(getdatestr(last_modified));

	fprintf(fp, ";\n");
	fprintf(fp, "; This zone information is generated by %s - do not edit!\n", prog_name);
	fprintf(fp, ";\n");
	fprintf(fp, "; $Id: edb_dns_make.c,v 1.4 2006/01/25 02:29:10 alex Exp alex $\n");
	fprintf(fp, ";\n");
	fprintf(fp, "$ORIGIN\t%s.", domainname);
	fprintf(fp, "\n");
	fprintf(fp, "$TTL\t%d\n", NS_MINIMUM);
	fprintf(fp, "\n");
	fprintf(fp, "@\tIN\tSOA\t%s. postmaster.%s. (\n", ns1, domainname);
	fprintf(fp, "\t\t\t%s\t; Serial\n", serial);
	fprintf(fp, "\t\t\t%d\t\t; Refresh %d hour\n", NS_REFRESH, NS_REFRESH/3600);	/* Refresh */
	fprintf(fp, "\t\t\t%d\t\t; Retry %d hour\n", NS_RETRY, NS_RETRY/3600);	/* Retry */
	fprintf(fp, "\t\t\t%d\t\t; Expire %d hour\n", NS_EXPIRE, NS_EXPIRE/3600);	/* Expire */
	fprintf(fp, "\t\t\t%d\t\t; Minimum %d hour\n", NS_MINIMUM, NS_MINIMUM/3600);	/* Minimum */
	fprintf(fp, "\t\t\t)\n");

	ca_ns = edb_dns_lookup_iphosts_by_TC(ec, tc_ns);
	ForCatalogueLink(l, ca_ns) {
		EdbTupleInfo	host = CatalogueObject(l);
		char	*fqdn;

		fqdn = edb_dns_host_fqdn(ec, host);
		str_minuscule1(fqdn);
		fprintf(fp, "\t\tIN\tNS\t%s.\n", fqdn);
		FREE(fqdn);
	}
	edb_catalogue_free(ca_ns);

	preference = 10;
	ca_mx_printed = edb_catalogue_new();
	ForTCDatum(datum, tc_mx) {
		char	*fqdn;
		EdbCatalogue	ca_mx;

		if(!DatumEIDisVALID(datum)) continue;
		ca_mx = edb_dns_lookup_iphosts_by_EID(ec, DatumEID(datum), ec->max_mapto_recurse);
		if(ca_mx) {
			Link	lm;

			ForCatalogueLink(lm, ca_mx) {
				host = CatalogueObject(lm);
				if(edb_catalogue_lookup(ca_mx_printed, TupleEID(host))) continue;
				fqdn = edb_dns_host_fqdn(ec, host);
				str_minuscule1(fqdn);
				fprintf(fp, "\t\tIN\tMX\t%d\t%s.\n", preference, fqdn);
				FREE(fqdn);
				edb_catalogue_append(ca_mx_printed, TupleEID(host), host);
			}
			edb_catalogue_free(ca_mx);
		}
		++preference;
	}
	edb_catalogue_free(ca_mx_printed);
#if 0	/* wide card MX is desabled (1998/0604) */
	ms = domain->field[ATON('r')];
	preference = 1000;
	{
		char *s;

		s = strdup(ms->d);
		str_minuscule1(s);
		fprintf(fp, "*\t\tIN\tMX\t%d\t%s.\n", preference, s);
		free(s);
	}
#endif

	fprintf(fp, ";\n");
	fprintf(fp, "; Host Information\n");
	fprintf(fp, ";\n");
	fprintf(fp, "$ORIGIN\t%s.\n", domainname);

	l_hosts = edb_dns_zone_get_hosts(ec, domain);
	ForLink(l, l_hosts) {
		EdbTupleInfo	host = l->d;

		edb_dns_host_to_zoneinfo(ec, domain, host, fp);
		edb_catalogue_append(ca_host_printed, TupleEID(host), host);
	}
	link_free(l_hosts);

	fprintf(fp, ";\n");
	fprintf(fp, "; Domain Information\n");
	fprintf(fp, ";\n");
	fprintf(fp, "$ORIGIN\t%s.\n", domainname);
	l_domains = edb_dns_zone_get_subdomains(ec, domain);
	ForLink(l, l_domains) {
		edb_dns_domain_to_zoneinfo(ec, domain, l->d, ca_host_printed, fp);
	}
	link_free(l_domains);

	edb_catalogue_free(ca_host_printed);

	FREE(domainname);
	FREE(serial);
	fprintf(fp, ";\n");
	fprintf(fp, "%s\n", END_OF_ZONEINFO);
	fprintf(fp, ";\n");
	return TRUE;
}

#if 0

static char *sorting_network;

/*#hosts_compare_by_ip--- hosts compare */
static
int
hosts_compare_by_ip(const void *v1, const void *v2)
{
	Wdb	*h1 = (Wdb *)v1;
	Wdb	*h2 = (Wdb *)v2;
	Link ip1, ip2;

	for(ip1 = (*h1)->field[ATON('b')]; ip1; ip1=ip1->n) {
		if(dns_check_addr_in_network(ip1->d, sorting_network)) break;
	}
	for(ip2 = (*h2)->field[ATON('b')]; ip2; ip2=ip2->n) {
		if(dns_check_addr_in_network(ip2->d, sorting_network)) break;
	}

	if(ip1 && ip2) {
		register in_addr_t	ia1 = inet_addr(ip1->d);
		register in_addr_t	ia2 = inet_addr(ip2->d);

		return ntohl(ia1) - ntohl(ia2);
	} else if(ip1) {
		return 1;
	} else if(ip2) {
		return -1;
	} else {
		return 0;
	}
}

/*#hosts_sort_by_ip--- sort list */
static
Link
hosts_sort_by_ip(char *network, Link hosts)
{
	int i, n;
	Wdb *s;
	Link l;

	sorting_network = strdup(network);

	n = link_atoms(hosts);
	s = (Wdb *)alloca(sizeof(Wdb)*n);
	if( s == NULL ) return NULL;
	for(i=0, l=hosts; i<n; ++i, l=l->n) s[i] = (Wdb)l->d;

	qsort(s, n, sizeof(Wdb), hosts_compare_by_ip);

	for(i=0, l=hosts; i<n; ++i, l=l->n) (Wdb)l->d = s[i];

	free(sorting_network);
	return hosts;
}
#endif 

static
int
compare_revinfo(LinkObj o1, LinkObj o2)
{
	RevInfo	ri1 = o1->d;
	RevInfo	ri2 = o2->d;

	if(ri1->addr<ri2->addr) return -1;
	if(ri1->addr>ri2->addr) return 1;
	return 0;
}

static
int
edb_dns_host_to_revinfo(EdbContext ec, EdbTupleInfo network, RevInfo rev, FILE *fp)
{
	Link ip, net, name;
	char	*fqdn;

	if(network==NULL) return FALSE;

	fqdn = edb_dns_host_fqdn(ec, rev->host);
	if(fqdn==NULL) return FALSE;
	str_minuscule1(fqdn);
	fprintf(fp, "%d\t\tIN\tPTR\t%s.\n", rev->addr&0xff, fqdn);
	FREE(fqdn);
	return TRUE;
}

static
Link
edb_dns_rev_get_hosts(EdbContext ec, EdbTupleInfo network)
{
	Link	l, l_hosts, l_tuples;
	EdbColumnInfo	ci;
	char	*query, *condition;
	char	*s, *name;
	eid_t	eid = 0;
	int	n;
	in_addr_t	netmask;
	in_addr_t	netaddr;
	EdbTupleColumn	tc;
	EdbDatum	datum;
	char	*cidr, *re;
	int	mask;

	l_tuples = edb_dns_lookup_hosts_by_network(ec, network);

	tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.address");
	if(tc==NULL || (datum=tc->datum)==NULL) return NULL;
	cidr = (char *)DatumEnglish(datum);
	if(!STRisVALID(cidr)) return NULL;
	cidr = STRDUP(cidr);
	netaddr	= cidr_to_address(cidr);
	mask	= cidr_to_mask(cidr);
	netmask	= ((~0)<<(32-mask));

	l_hosts = NULL;
	ForLink(l, l_tuples) {
		EdbTupleInfo	tuple = l->d;
		EdbTupleColumn	tc;
		EdbDatum	datum;

		if(!tuple) continue;
		tc = edb_tuple_seek_column_by_xmlname(ec, tuple, (utf8_t *)"@.address");
		ForTCDatum(datum, tc) {
			RevInfo	rev;
			in_addr_t	addr;

			if(!DatumEnglishIsValid(datum)) continue;
			addr = inet_addr((char *)DatumEnglish(datum));
			addr = ntohl(addr);
			if((addr&netmask)!=netaddr) continue;

			rev = (RevInfo)malloc(sizeof(struct REV_INFO));
			rev->addr = addr;
			rev->host = tuple;
			LinkAppend(l_hosts, link_atom_new(rev, free));
		}
	}

	link_free(l_tuples);
	return l_hosts;
}

static
Link
edb_dns_rev_get_network16to24(EdbContext ec, EdbTupleInfo network)
{
	EdbTableInfo	ti;
	char	*query, *condition;
	char	*s, *name;
	Link	l, l_eid, l_network;
	eid_t	eid = 0;
	int	n;
	in_addr_t	netaddr;
	EdbTupleColumn	tc;
	EdbDatum	datum;
	char	*cidr, *re;
	int	mask;

	tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.address");
	if(tc==NULL || (datum=tc->datum)==NULL) return NULL;
	cidr = (char *)DatumEnglish(datum);
	if(!STRisVALID(cidr)) return NULL;
	cidr = STRDUP(cidr);
	netaddr	= cidr_to_address(cidr);
	mask	= cidr_to_mask(cidr);
	asprintf(&re, "^%d[.]%d[.].*/24$", (netaddr>>24)&0xff, (netaddr>>16)&0xff);

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetNetwork_XN);
	if(ti==NULL) {
		edb_trace(ec, "ERROR: edb_dns_rev_get_network16to24: can't get %s\n", EDB_TABLE_InetNetwork_XN);
		return NULL;
	}
	asprintf(&query, "SELECT " EDB_BASE_EID " FROM %s WHERE %s ORDER BY %s::inet," EDB_BASE_EID,
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionRegexTextEQ("@.address", re),
			ConditionTerminate),
			ColumnSN(tc->ci));
	FREE(condition);
	FREE(re);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_network = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	tuple = edb_tuple_get(ec, ATOI(l->d), 0);

		if(!tuple) continue;
		LinkAppend(l_network, link_atom_new(tuple, NULL));
	}
	link_free(l_eid);
	return l_network;
}

static
int
edb_dns_make_revinfo(EdbContext ec, char *cidr, FILE *fp)
{
	Link network_elm, l;
	char last_modified[BUFSIZ];
	char *serial, *ns1;
	EdbTupleInfo	network, host;
	EdbTupleColumn	tc_ns;
	EdbCatalogue		ca_ns;
	EdbDatum	d;
	int	mask;
	in_addr_t	netaddr;

	netaddr = cidr_to_address(cidr);
	mask = cidr_to_mask(cidr);
	network = edb_dns_lookup_network_by_cidr(ec, cidr);
	if(network==NULL) {
		fprintf(fp, "; can't get network information %s\n", cidr);
		return FALSE;
	}

	tc_ns = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.ns");
	if(tc_ns==NULL) {
		fprintf(fp, "; No NS  is defined in network %s\n", cidr);
		return FALSE;
	}
	ns1 = NULL;
	if(tc_ns && (d=tc_ns->datum) && DatumEIDisVALID(d)) {
		host = edb_tuple_get(ec, DatumEID(d), 0);
		if(host) ns1 = edb_dns_host_fqdn(ec, host);
	}
	if(!STRisVALID(ns1)) {
		fprintf(fp, "; No NS is defined in network %s\n", cidr);
		return FALSE;
	}
	network_elm = str_split(cidr, '.');
	network_elm = link_reverse(network_elm);
	str_minuscule1(ns1);
	serial = time_to_serial(getdatestr(last_modified));

	fprintf(fp, ";\n");
	fprintf(fp, "; This reverse information is generated by %s - do not edit!\n", prog_name);
	fprintf(fp, ";\n");
	fprintf(fp, "; $Id: edb_dns_make.c,v 1.4 2006/01/25 02:29:10 alex Exp alex $\n");
	fprintf(fp, ";\n");
	if(mask==16) {
		fprintf(fp, "$ORIGIN\t%s.%s.in-addr.arpa.\n", network_elm->n->n->d, network_elm->n->n->n->d);
	} else {
		fprintf(fp, "$ORIGIN\t%s.%s.%s.in-addr.arpa.\n", network_elm->n->d, network_elm->n->n->d, network_elm->n->n->n->d);
	}
	fprintf(fp, "\n");
	fprintf(fp, "$TTL\t%d\n", NS_MINIMUM);
	fprintf(fp, "\n");
	fprintf(fp, "@\tIN\tSOA\t%s. postmaster.%s. (\n", ns1, ns1);
	fprintf(fp, "\t\t\t%s\t; Serial\n", serial);
	fprintf(fp, "\t\t\t%d\t\t; Refresh %d hour\n", NS_REFRESH, NS_REFRESH/3600);	/* Refresh */
	fprintf(fp, "\t\t\t%d\t\t; Retry %d hour\n", NS_RETRY, NS_RETRY/3600);	/* Retry */
	fprintf(fp, "\t\t\t%d\t\t; Expire %d hour\n", NS_EXPIRE, NS_EXPIRE/3600);	/* Expire */
	fprintf(fp, "\t\t\t%d\t\t; Minimum %d hour\n", NS_MINIMUM, NS_MINIMUM/3600);	/* Minimum */
	fprintf(fp, "\t\t\t)\n");

	ca_ns = edb_dns_lookup_iphosts_by_TC(ec, tc_ns);
	ForCatalogueLink(l, ca_ns) {
		EdbTupleInfo	host = CatalogueObject(l);
		char	*fqdn;

		fqdn = edb_dns_host_fqdn(ec, host);
		str_minuscule1(fqdn);
		fprintf(fp, "\t\tIN\tNS\t%s.\n", fqdn);
		FREE(fqdn);
	}
	edb_catalogue_free(ca_ns);

	fprintf(fp, ";\n");
	fprintf(fp, "; Host Information\n");
	fprintf(fp, ";\n");
	if(mask==16) {
		fprintf(fp, "$ORIGIN\t%s.%s.in-addr.arpa.\n", network_elm->n->n->d, network_elm->n->n->n->d);
	} else {
		fprintf(fp, "$ORIGIN\t%s.%s.%s.in-addr.arpa.\n", network_elm->n->d, network_elm->n->n->d, network_elm->n->n->n->d);
	}

	if(mask==16) {
		Link	l_net;

		l_net = edb_dns_rev_get_network16to24(ec, network);
		ForLink(l, l_net) {
			EdbTupleInfo	cnet = l->d;
			char	*ccidr;
			in_addr_t	caddress;

			if(cnet==NULL) continue;
			ccidr = edb_dns_network_cidr(ec, cnet);
			caddress = cidr_to_address(ccidr);
			FREE(ccidr);
			tc_ns = edb_tuple_seek_column_by_xmlname(ec, cnet, (utf8_t *)"@.ns");
			if(tc_ns) {
				EdbCatalogue	ca_ns = edb_dns_lookup_iphosts_by_TC(ec, tc_ns);
				Link	ln;
				char	*s, *caddr;

				asprintf(&s, "%d", (caddress>>8)&0xff);
				caddr = STRDUP(s);
				FREE(s);
				ForCatalogueLink(ln, ca_ns) {
					EdbTupleInfo	ns = CatalogueObject(ln);
					char	*cfqdn = NULL;

					if(ns && (cfqdn = edb_dns_host_fqdn(ec, ns))) {
						str_minuscule1(cfqdn);
						fprintf(fp, "%s\t\tIN\tNS\t%s.\n", caddr, cfqdn);
						caddr = "";
					}
					FREE(cfqdn);
				}
				edb_catalogue_free(ca_ns);
			}
		}
	} else if(mask==24) {
		Link	l_rev;

		l_rev = edb_dns_rev_get_hosts(ec, network);
		l_rev = link_sort(l_rev, compare_revinfo, NULL, NULL, NULL);
		ForLink(l, l_rev) {
			edb_dns_host_to_revinfo(ec, network, l->d, fp);
		}
		link_free(l_rev);
	}

	link_free(network_elm);
	FREE(serial);
	fprintf(fp, ";\n");
	fprintf(fp, "%s\n", END_OF_REVINFO);
	fprintf(fp, ";\n");
	return TRUE;
}

#if 0

static Link wdb_look();
static Link wdb_look_relation();

static int list_compare();
static Link list_sort();

static Link find_list;

#include	<signal.h>

/*#suppress_identical--- suppress identical list */
static
void
suppress_identical(list)
Link list;
{

	while(list) {
		if(list->n && !strcmp(list->d, list->n->d)) {
			Link t;

			t = list->n;
			list->n = list->n->n;
			if(t->n) t->n->p = list;
			link_atom_free(t);
		} else {
			list = list->n;
		}
	}
}

/*#wdb_look--- look whois data base */
static
Link
wdb_look(section, target, mode)
struct SECTION *section;
char *target;
int mode;
{
	Link	list;
	char c, *sp;

	list = NULL;
	sp = section->search;
	while((c = *sp++)) {
		LinkAppend(list, wdb_index_search(section->type, c, target, mode));
	}
	return list;
}

/*#wdb_look_relation--- look whois data base */
static
Link
wdb_look_relation(section, target, mode)
struct SECTION *section;
char *target;
int mode;
{
	Link	list;
	char c, *sp;

	list = NULL;
	sp = section->relation;
	while((c = *sp++)) {
		LinkAppend(list, wdb_index_search(section->type, c, target, mode));
	}
	return list;
}

/*#list_compare--- list compare */
static
int
list_compare(s1, s2)
char **s1;
char **s2;
{

	return strcmp(*s1, *s2);
}

/*#list_sort--- sort list */
static
Link
list_sort(list)
Link list;
{
	int i, n;
	char **s;
	Link l;

	n = link_atoms(list);
	s = (char **)alloca(sizeof(char *)*n);
	if( s == NULL ) return NULL;
	for(i=0, l=list; i<n; ++i, l=l->n) s[i] = (char *)l->d;

	qsort(s, n, sizeof(char *), list_compare);

	for(i=0, l=list; i<n; ++i, l=l->n) (char *)l->d = s[i];

	return list;
}
#endif

/*#print_file--- print help message */
static
void
print_file(char *name)
{
	FILE *fp;
	char buf[BUFSIZ];

	fp = fopen(name, "r");
	if(fp == NULL) return;
	while(!feof(fp)) {
		if(fgets(buf, sizeof(buf), fp) != NULL) {
			printf("%s", buf);
		}
	}
	fclose(fp);
}

/*#message_print--- message print */
static
int
message_print(char *name, FILE *out)
{
	FILE *fp;
	char buf[BUFSIZ];

	fp = fopen(name, "r");
	if(fp == NULL) return 0;
	while(!feof(fp)) {
		if(!fgets(buf, sizeof(buf), fp)) break;
		fprintf(out, "%s", buf);
	}
	fclose(fp);
	return 1;
}

static
struct LABEL wdb_domain_label[FIELDs] = {
{"[ドメイン名]",	"[Domain]",			DOMAIN,		STAFF},	/* a */
{NULL,			NULL,				-1,			NODEF},	/* b */
{NULL,			NULL,				-1,			NODEF},	/* c */
{NULL,			NULL,				-1,			NODEF},	/* d */
{NULL,			NULL,				-1,			NODEF},	/* e */
{"[組織名]",		NULL,				-1,			ANY},	/* f */
{"[Organization]",	"[Organization]",		-1,			ANY},	/* g */
{"[郵便番号]",		"[ZIP]",				-1,			ANY},	/* h */
{"[住所]",			NULL,				-1,			ANY},	/* i */
{"[Address]",		"[Address]",			-1,			ANY},	/* j */
{NULL,			NULL,				-1,			NODEF},	/* k */
{NULL,			NULL,				-1,			NODEF},	/* l */
{NULL,			NULL,				-1,			NODEF},	/* m */
{"[運用責任者]",	"[Administrative Contact]",	PERSONAL,	ANY},	/* n */
{"[技術担当者]",	"[Technical Contact]",	PERSONAL,	ANY},	/* o */
{"[ネームサーバ]",	"[Name Server]",		HOST,		STAFF},	/* p */
{NULL,			NULL,				-1,			NODEF},	/* q */
{"[メイルサーバ]",	"[Mail Server]",			HOST,		STAFF},	/* r */
{NULL,			NULL,				-1,			NODEF},	/* s */
{"[IPネットワーク]",	"[IP Network]",			NETWORK,	STAFF},	/* t */
{"[URL]",			"[URL]",				-1,			ANY},	/* u */
{NULL,			NULL,				-1,			NODEF},	/* v */
{NULL,			NULL,				-1,			NODEF},	/* w */
{NULL,			NULL,				-1,			NODEF},	/* x */
{NULL,			NULL,				-1,			NODEF},	/* y */
{NULL,			NULL,				-1,			NODEF}	/* z */
};

struct SECTION wdb_domain_section = {
	DOMAIN_TYPE,
	"Domain Information:",
	"[ドメイン情報]",
	wdb_domain_label,
	"afg",
	"noprt",
};

static
int
edb_wdb_domain_print(EdbContext ec, EdbTupleInfo domain, int japanese, int mode, Doc doc)
{
	EdbTupleColumn	tc;
	char	*fqdn;
	struct SECTION	*section = &wdb_domain_section;
	struct LABEL		*label = section->label;
	EdbDatum	datum;

	if((mode&PRINT_LONGFORM)==0) {
		EdbDatum	datum;

		tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.organization");
		if(tc && (datum=tc->datum)) {
			if(japanese && DatumJapaneseIsValid(datum)) {
				doc_text_printf(doc, "%s", (char *)DatumJapanese(datum));
			} else if(DatumEnglishIsValid(datum)) {
				doc_text_printf(doc, "%s", (char *)DatumEnglish(datum));
			}
		}

		fqdn = edb_dns_domain_fqdn(ec, domain);
		if(STRisVALID(fqdn)) doc_text_printf(doc, " (%s)", fqdn);
		FREE(fqdn);

		doc_text_printf(doc, "\tEID=%d\n", TupleEID(domain));
		return TRUE;
	}

	doc_text_puts(doc, section->name);
	doc_puts(doc, " ");
	doc_text_puts(doc, section->jname);
	doc_puts(doc, "\n");

	fqdn = edb_dns_domain_fqdn(ec, domain);
	if(STRisVALID(fqdn)) doc_text_printf(doc, "a. %s\t%s\n", BILABEL(label, 'a', japanese), fqdn);
	FREE(fqdn);

	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.organization");
	ForTCDatum(datum, tc) {
		if(japanese && DatumJapaneseIsValid(datum)) {
			doc_text_printf(doc, "f. %s\t%s\n", BILABEL(label, 'f', japanese), DatumJapanese(datum));
		}
		if(DatumEnglishIsValid(datum)) {
			doc_text_printf(doc, "g. %s\t%s\n", BILABEL(label, 'g', japanese), DatumEnglish(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.administrator");
	ForTCDatum(datum, tc) {
		if(DatumEIDisVALID(datum)) {
			doc_text_printf(doc, "n. %s\tEID=%d\n", BILABEL(label, 'n', japanese), DatumEID(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.contact");
	ForTCDatum(datum, tc) {
		if(DatumEIDisVALID(datum)) {
			doc_text_printf(doc, "o. %s\tEID=%d\n", BILABEL(label, 'o', japanese), DatumEID(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.ns");
	ForTCDatum(datum, tc) {
		if(DatumEIDisVALID(datum)) {
			EdbTupleInfo	ns = edb_tuple_get(ec, DatumEID(datum), 0);

			if(ns) {
				fqdn = edb_dns_host_fqdn(ec, ns);
				str_minuscule1(fqdn);
				if(STRisVALID(fqdn)) doc_text_printf(doc, "p. %s\t%s\n", BILABEL(label, 'p', japanese), fqdn);
				FREE(fqdn);
			}
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.mx");
	ForTCDatum(datum, tc) {
		if(DatumEIDisVALID(datum)) {
			EdbTupleInfo	mx = edb_tuple_get(ec, DatumEID(datum), 0);

			if(mx) {
				fqdn = edb_dns_host_fqdn(ec, mx);
				str_minuscule1(fqdn);
				if(STRisVALID(fqdn)) doc_text_printf(doc, "r. %s\t%s\n", BILABEL(label, 'r', japanese), fqdn);
				FREE(fqdn);
			}
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.ipnetwork");
	ForTCDatum(datum, tc) {
		if(DatumEIDisVALID(datum)) {
			EdbTupleInfo	net = edb_tuple_get(ec, DatumEID(datum), 0);
			char	*cidr;

			if(net) {
				cidr = edb_dns_network_cidr(ec, net);
				if(STRisVALID(cidr)) doc_text_printf(doc, "t. %s\t%s\n", BILABEL(label, 't', japanese), cidr);
				FREE(cidr);
			}
		}
	}

	return TRUE;
}

static
struct LABEL wdb_network_label[FIELDs] = {
{"[IPネットワーク]",		"[IP Network]",			NETWORK	,	STAFF},	/* a */
{"[ネットワーク省略名]",	"[Network Name]",		-1,			STAFF},	/* b */
{NULL,				NULL,				-1,			NODEF},	/* c */
{NULL,				NULL,				-1,			NODEF},	/* d */
{NULL,				NULL,				-1,			NODEF},	/* e */
{"[組織名]",			NULL,				-1,			ANY},	/* f */
{"[Organization]",		"[Organization]",		-1,			ANY},	/* g */
{"[ZIP]",				"[ZIP]",				-1,			ANY},	/* h */
{"[住所]",				NULL,				-1,			ANY},	/* i */
{"[Address]",			"[Address]",			-1,			ANY},	/* j */
{NULL,				NULL,				-1,			NODEF},	/* k */
{NULL,				NULL,				-1,			NODEF},	/* l */
{NULL,				NULL,				-1,			NODEF},	/* m */
{"[運用責任者]",		"[Administrative Contact]",	PERSONAL,	ANY},	/* n */
{"[技術担当者]",		"[Technical Contact]",	PERSONAL,	ANY},	/* o */
{"[ネームサーバ]",		"[Name Server]",		HOST,		STAFF},	/* p */
{NULL,				NULL,				-1,			NODEF},	/* q */
{NULL,				NULL,				-1,			NODEF},	/* r */
{NULL,				NULL,				-1,			NODEF},	/* s */
{NULL,				NULL,				-1,			NODEF},	/* t */
{"[URL]",				"[URL]",				-1,			ANY},	/* u */
{NULL,				NULL,				-1,			NODEF},	/* v */
{NULL,				NULL,				-1,			NODEF},	/* w */
{NULL,				NULL,				-1,			NODEF},	/* x */
{NULL,				NULL,				-1,			NODEF},	/* y */
{NULL,				NULL,				-1,			NODEF}	/* z */
};

struct SECTION wdb_network_section = {
	NETWORK_TYPE,
	"Network Information:",
	"[ネットワーク情報]",
	wdb_network_label,
	"abfg",
	"nop",
};


static
int
edb_wdb_network_print(EdbContext ec, EdbTupleInfo network, int japanese, int mode, Doc doc)
{
	EdbTupleColumn	tc;
	char	*cidr, *fqdn;
	struct SECTION	*section = &wdb_network_section;
	struct LABEL		*label = section->label;
	EdbDatum	datum;

	if((mode&PRINT_LONGFORM)==0) {
		EdbDatum	datum;

		tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.organization");
		if(tc && (datum=tc->datum)) {
			if(japanese && DatumJapaneseIsValid(datum)) {
				doc_text_printf(doc, "%s", DatumJapanese(datum));
			} else if(DatumEnglishIsValid(datum)) {
				doc_text_printf(doc, "%s", DatumEnglish(datum));
			}
		}

		tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.name");
		if(tc && (datum=tc->datum)) {
			if(DatumEnglishIsValid(datum)) {
				doc_text_printf(doc, " (%s)", DatumEnglish(datum));
			}
		}

		cidr = edb_dns_network_cidr(ec, network);
		if(STRisVALID(cidr)) doc_text_printf(doc, " (%s)", cidr);
		FREE(cidr);

		doc_text_printf(doc, "\tEID=%d\n", TupleEID(network));
		return TRUE;
	}

	doc_text_puts(doc, section->name);
	doc_puts(doc, " ");
	doc_text_puts(doc, section->jname);
	doc_puts(doc, "\n");

	cidr = edb_dns_network_cidr(ec, network);
	if(STRisVALID(cidr)) doc_text_printf(doc, "a. %s\t%s\n", BILABEL(label, 'a', japanese), cidr);
	FREE(cidr);

	tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.name");
	ForTCDatum(datum, tc) {
		if(DatumEnglishIsValid(datum)) {
			doc_text_printf(doc, "b. %s\t%s\n", BILABEL(label, 'b', japanese), DatumEnglish(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.organization");
	ForTCDatum(datum, tc) {
		if(japanese && DatumJapaneseIsValid(datum)) {
			doc_text_printf(doc, "f. %s\t%s\n", BILABEL(label, 'f', japanese), DatumJapanese(datum));
		}
		if(DatumEnglishIsValid(datum)) {
			doc_text_printf(doc, "g. %s\t%s\n", BILABEL(label, 'g', japanese), DatumEnglish(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.administrator");
	ForTCDatum(datum, tc) {
		if(DatumEIDisVALID(datum)) {
			doc_text_printf(doc, "n. %s\tEID=%d\n", BILABEL(label, 'n', japanese), DatumEID(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.contact");
	ForTCDatum(datum, tc) {
		if(DatumEIDisVALID(datum)) {
			doc_text_printf(doc, "o. %s\tEID=%d\n", BILABEL(label, 'o', japanese), DatumEID(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.ns");
	ForTCDatum(datum, tc) {
		if(DatumEIDisVALID(datum)) {
			EdbTupleInfo	ns = edb_tuple_get(ec, DatumEID(datum), 0);

			if(ns) {
				fqdn = edb_dns_host_fqdn(ec, ns);
				str_minuscule1(fqdn);
				if(STRisVALID(fqdn)) doc_text_printf(doc, "p. %s\t%s\n", BILABEL(label, 'p', japanese), fqdn);
				FREE(fqdn);
			}
		}
	}

	return TRUE;
}

static
struct LABEL wdb_host_label[FIELDs] = {
{"[ホスト名]",		"[Host Name]",			HOST,		ANY},	/* a */
{"[IP Address]",		"[IP Address]",			-1,			ANY},	/* b */
{"[CPU]",			"[CPU]",				-1,			ANY},	/* c */
{"[OS]",			"[OS]",				-1,			ANY},	/* d */
{NULL,			NULL,				-1,			NODEF},	/* e */
{"[別称]",			"[Aliased Name]",		-1,			ANY},	/* f */
{NULL,			NULL,				-1,			NODEF},	/* g */
{NULL,			NULL,				-1,			NODEF},	/* h */
{NULL,			NULL,				-1,			NODEF},	/* i */
{NULL,			NULL,				-1,			NODEF},	/* j */
{NULL,			NULL,				-1,			NODEF},	/* k */
{"[設置場所]",		NULL,				-1,			ANY},	/* l */
{NULL,			NULL,				-1,			NODEF},	/* m */
{"[運用責任者]",	"[Administrative Contact]",	PERSONAL,	ANY},	/* n */
{"[技術担当者]",	"[Technical Contact]",	PERSONAL,	ANY},	/* o */
{NULL,			NULL,				-1,			NODEF},	/* p */
{NULL,			NULL,				-1,			NODEF},	/* q */
{"[メイルサーバ]",	"[Mail Server]",			HOST,		ANY},	/* r */
{NULL,			NULL,				-1,			NODEF},	/* s */
{NULL,			NULL,				-1,			NODEF},	/* t */
{"[URL]",			"[URL]",				-1,			ANY},	/* u */
{NULL,			NULL,				-1,			NODEF},	/* v */
{NULL,			NULL,				-1,			NODEF},	/* w */
{NULL,			NULL,				-1,			NODEF},	/* x */
{NULL,			NULL,				-1,			NODEF},	/* y */
{NULL,			NULL,				-1,			NODEF}	/* z */
};

struct SECTION wdb_host_section = {
	HOST_TYPE,
	"Host Information:",
	"[ホスト情報]",
	wdb_host_label,
	"abf",
	"nor",
};

static
int
edb_wdb_host_print(EdbContext ec, EdbTupleInfo host, int japanese, int mode, Doc doc)
{
	EdbTupleColumn	tc;
	char	*fqdn;
	struct SECTION	*section = &wdb_host_section;
	struct LABEL		*label = section->label;
	EdbDatum	datum;

	if((mode&PRINT_LONGFORM)==0) {
		EdbDatum	datum;

		fqdn = edb_dns_host_fqdn(ec, host);
		if(STRisVALID(fqdn)) doc_text_printf(doc, "%s", fqdn);
		FREE(fqdn);

		tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.address");
		ForTCDatum(datum, tc) {
			if(DatumEnglishIsValid(datum)) doc_text_printf(doc, " (%s)", DatumEnglish(datum));
		}

		doc_text_printf(doc, "\tEID=%d\n", TupleEID(host));
		return TRUE;
	}

	doc_text_puts(doc, section->name);
	doc_puts(doc, " ");
	doc_text_puts(doc, section->jname);
	doc_puts(doc, "\n");

	fqdn = edb_dns_host_fqdn(ec, host);
	if(STRisVALID(fqdn)) doc_text_printf(doc, "a. %s\t%s\n", BILABEL(label, 'a', japanese), fqdn);
	FREE(fqdn);

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.address");
	ForTCDatum(datum, tc) {
		if(DatumEnglishIsValid(datum)) doc_text_printf(doc, "b. %s\t%s\n", BILABEL(label, 'b', japanese), DatumEnglish(datum));
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.cpu");
	ForTCDatum(datum, tc) {
		if(japanese && DatumJapaneseIsValid(datum)) {
			doc_text_printf(doc, "c. %s\t%s\n", BILABEL(label, 'c', japanese), DatumJapanese(datum));
		} else if(DatumEnglishIsValid(datum)) {
			doc_text_printf(doc, "c. %s\t%s\n", BILABEL(label, 'c', japanese), DatumEnglish(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.os");
	ForTCDatum(datum, tc) {
		if(japanese && DatumJapaneseIsValid(datum)) {
			doc_text_printf(doc, "d. %s\t%s\n", BILABEL(label, 'd', japanese), DatumJapanese(datum));
		} else if(DatumEnglishIsValid(datum)) {
			doc_text_printf(doc, "d. %s\t%s\n", BILABEL(label, 'd', japanese), DatumEnglish(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.location");
	ForTCDatum(datum, tc) {
		if(japanese && DatumJapaneseIsValid(datum)) {
			doc_text_printf(doc, "l. %s\t%s\n", BILABEL(label, 'l', japanese), DatumJapanese(datum));
		} else if(DatumEnglishIsValid(datum)) {
			doc_text_printf(doc, "l. %s\t%s\n", BILABEL(label, 'l', japanese), DatumEnglish(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.administrator");
	ForTCDatum(datum, tc) {
		if(DatumEIDisVALID(datum)) {
			doc_text_printf(doc, "n. %s\tEID=%d\n", BILABEL(label, 'n', japanese), DatumEID(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.contact");
	ForTCDatum(datum, tc) {
		if(DatumEIDisVALID(datum)) {
			doc_text_printf(doc, "o. %s\tEID=%d\n", BILABEL(label, 'o', japanese), DatumEID(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.mx");
	ForTCDatum(datum, tc) {
		if(DatumEIDisVALID(datum)) {
			EdbTupleInfo	mx = edb_tuple_get(ec, DatumEID(datum), 0);

			if(mx) {
				fqdn = edb_dns_host_fqdn(ec, mx);
				str_minuscule1(fqdn);
				if(STRisVALID(fqdn)) doc_text_printf(doc, "r. %s\t%s\n", BILABEL(label, 'r', japanese), fqdn);
				FREE(fqdn);
			}
		}
	}

	return TRUE;
}

static
struct LABEL wdb_person_label[FIELDs] = {
{"[ハンドル]",		"[Handle]",			PERSONAL,	ANY},		/* a */
{"[氏名]",			NULL,			-1,			ANY},		/* b */
{"[Last, First]",		"[Last, First]",		-1,			ANY},		/* c */
{"[電子メイル]",	"[E-mail]",			-1,			ANY},		/* d */
{"[JPNICハンドル]",	"[JPNIC Handle]",	-1,			ANY},		/* e */
{"[組織名]",		NULL,			-1,			ANY},		/* f */
{"[Organization]",	"[Organization]",	-1,			ANY},		/* g */
{"[郵便番号]",		"[ZIP]",			-1,			ANY},		/* h */
{"[住所]",			NULL,			-1,			ANY},		/* i */
{"[Address]",		"[Address]",		-1,			ANY},		/* j */
{"[部局名]",		NULL,			-1,			ANY},		/* k */
{"[Division]",		"[Division]",		-1,			ANY},		/* l */
{"[肩書]",			NULL,			-1,			ANY},		/* m */
{"[Title]",			"[Title]",			-1,			ANY},		/* n */
{"[電話番号]",		"[Phone]",			-1,			ANY},		/* o */
{"[FAX番号]",		"[Fax]",			-1,			ANY},		/* p */
{NULL,			NULL,			-1,			NODEF},		/* q */
{NULL,			NULL,			-1,			NODEF},		/* r */
{NULL,			NULL,			-1,			NODEF},		/* s */
{NULL,			NULL,			-1,			NODEF},		/* t */
{"[URL]",			"[URL]",			-1,			ANY},		/* u */
{NULL,			NULL,			-1,			NODEF},		/* v */
{NULL,			NULL,			-1,			NODEF},		/* w */
{NULL,			NULL,			-1,			NODEF},		/* x */
{NULL,			NULL,			-1,			NODEF},		/* y */
{NULL,			NULL,			-1,			NODEF}		/* z */
};

struct SECTION wdb_person_section = {
	PERSONAL_TYPE,
	"Personal Information:",
	"[個人情報]",
	wdb_person_label,
	"abcde",
	"",
};

static
int
edb_wdb_person_print(EdbContext ec, EdbTupleInfo person, int japanese, int mode, Doc doc)
{
	EdbTupleColumn	tc;
	char	*email;
	struct SECTION	*section = &wdb_person_section;
	struct LABEL		*label = section->label;
	Caption	cap;
	EdbDatum	datum;

	if((mode&PRINT_LONGFORM)==0) {
		cap = edb_tuple_make_caption(ec, person, CAP_long|CAP_norecurse);
		if(cap) {
			if(STRisVALID(cap->c_main)) doc_text_printf(doc, "%s", cap->c_main);
			if(STRisVALID(cap->c_sub)) doc_text_printf(doc, " (%s)", cap->c_sub);
			doc_text_printf(doc, "\t\tEID=%d\n", TupleEID(person));
			edb_caption_free(ec, cap);
		}
		return TRUE;
	}

	doc_text_puts(doc, section->name);
	doc_puts(doc, " ");
	doc_text_puts(doc, section->jname);
	doc_puts(doc, "\n");

	doc_text_printf(doc, "a. %s\tEID=%d\n", BILABEL(label, 'a', japanese), TupleEID(person));

	cap = edb_tuple_make_caption(ec, person, CAP_long|CAP_norecurse);
	if(STRisVALID(cap->c_main_japanese)) doc_text_printf(doc, "b. %s\t%s\n", BILABEL(label, 'b', japanese), cap->c_main_japanese);
	if(STRisVALID(cap->c_main_english)) doc_text_printf(doc, "c. %s\t%s\n", BILABEL(label, 'c', japanese), cap->c_main_english);

	email = (char *)edb_person_email(ec, person);
	if(STRisVALID(email)) doc_text_printf(doc, "d. %s\t%s\n", BILABEL(label, 'd', japanese), email);
	FREE(email);

	tc = edb_tuple_seek_column_by_xmlname(ec, person, (utf8_t *)"@.address");
	ForTCDatum(datum, tc) {
		if(!edb_datum_at_present(ec, datum)) continue;
		if(DatumEnglishIsValid(datum)) doc_text_printf(doc, "b. %s\t%s\n", BILABEL(label, 'b', japanese), DatumEnglish(datum));
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, person, (utf8_t *)"@.affiliation");
	ForTCDatum(datum, tc) {
		if(!edb_datum_at_present(ec, datum)) continue;
		if(DatumEIDisVALID(datum)) {
			cap = edb_make_caption(ec, DatumEID(datum), 0);
			if(cap && japanese && STRisVALID(cap->c_main_japanese)) {
				doc_text_printf(doc, "k. %s\t%s\n", BILABEL(label, 'k', japanese), cap->c_main_japanese);
			}
			if(cap && STRisVALID(cap->c_main_english)) {
				doc_text_printf(doc, "l. %s\t%s\n", BILABEL(label, 'l', japanese), cap->c_main_english);
			}
			edb_caption_free(ec, cap);
		} else {
			if(japanese && DatumJapaneseIsValid(datum)) {
				doc_text_printf(doc, "k. %s\t%s\n", BILABEL(label, 'k', japanese), DatumJapanese(datum));
			}
			if(DatumEnglishIsValid(datum)) {
				doc_text_printf(doc, "l. %s\t%s\n", BILABEL(label, 'l', japanese), DatumEnglish(datum));
			}
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, person, (utf8_t *)"@.title");
	ForTCDatum(datum, tc) {
		if(!edb_datum_at_present(ec, datum)) continue;
		if(japanese && DatumJapaneseIsValid(datum)) {
			doc_text_printf(doc, "m. %s\t%s\n", BILABEL(label, 'm', japanese), DatumJapanese(datum));
		}
		if(DatumEnglishIsValid(datum)) {
			doc_text_printf(doc, "n. %s\t%s\n", BILABEL(label, 'n', japanese), DatumEnglish(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, person, (utf8_t *)"@.phone");
	ForTCDatum(datum, tc) {
		if(!edb_datum_at_present(ec, datum)) continue;
		if(DatumEnglishIsValid(datum)) {
			doc_text_printf(doc, "o. %s\t%s\n", BILABEL(label, 'o', japanese), DatumEnglish(datum));
		}
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, person, (utf8_t *)"@.fax");
	ForTCDatum(datum, tc) {
		if(!edb_datum_at_present(ec, datum)) continue;
		if(DatumEnglishIsValid(datum)) {
			doc_text_printf(doc, "p. %s\t%s\n", BILABEL(label, 'p', japanese), DatumEnglish(datum));
		}
	}

	return TRUE;
}

static
int
edb_wdb_personification_print(EdbContext ec, EdbTupleInfo tuple, int lang, int mode, Doc doc)
{
	Caption	cap = edb_tuple_make_caption(ec, tuple, CAP_long|CAP_norecurse);

	if(cap) {
		if(STRisVALID(cap->c_main)) doc_text_printf(doc, "%s", cap->c_main);
		if(STRisVALID(cap->c_sub)) doc_text_printf(doc, " (%s)", cap->c_sub);
		doc_text_printf(doc, "\t\tEID=%d\n", TupleEID(tuple));
		edb_caption_free(ec, cap);
	}
	return TRUE;
}

static
int
edb_wdb_print(EdbContext ec, EdbTupleInfo tuple, int japanese, int mode, Doc doc)
{

	if(TupleIsInetDomain(tuple)) {
		return edb_wdb_domain_print(ec, tuple, japanese, mode, doc);
	} else if(TupleIsInetNetwork(tuple)) {
		return edb_wdb_network_print(ec, tuple, japanese, mode, doc);
	} else if(TupleIsInetHost(tuple)) {
		return edb_wdb_host_print(ec, tuple, japanese, mode, doc);
	} else if(TupleIsPerson(tuple)) {
		return edb_wdb_person_print(ec, tuple, japanese, mode, doc);
	} else if(TupleIsPersonification(tuple)) {
		return edb_wdb_personification_print(ec, tuple, japanese, mode, doc);
	}
	return FALSE;
}

struct SECTION *sections[] = {
	&wdb_domain_section,
	&wdb_network_section,
	&wdb_host_section,
	&wdb_person_section,
};

static
int
dns_look(EdbContext ec, int ac, char *av[])
{
	char *target;
	int i;
	int n;
	int sec;
	char *sectype;
	int japanese;
	Link list;
	EdbCatalogue	ca_list;
	int print_mode;
	int find_mode;
	int relation;
	Doc	doc;

	message_print(MESSAGE_DIR "/dns-opening", stdout);
	japanese = TRUE;
	n = ac;
	for(i=0; i<n; ++i) {
		if(!strcmp(*av, "help")) {
			print_file(HELP_FILE);
			exit(0);
		} else if(!strcmp(*av, "-e")) {
			japanese = FALSE;
			--ac;
			++av;
		} else if(!strcmp(*av, "-j")) {
			japanese = TRUE;
			--ac;
			++av;
		} else {
			break;
		}
	}
	if(ac==0) {
		printf("usage: %s string\n", prog_name);
		exit(1);
	} else if(ac==1) {
		sectype = NULL;
		target = strdup(av[0]);
	} else if(ac==2) {
		if(!strcmp(av[0], "DOM") || !strcmp(av[0], "NET") || !strcmp(av[0], "HOST") || !strcmp(av[0], "PERSON")) {
			sectype = strdup(av[0]);
			target = strdup(av[1]);
		} else {
			sectype= NULL;
			asprintf(&target, "%s/%s", av[0], av[1]);
		}
	} else {
		if(!strcmp(av[0], "DOM") || !strcmp(av[0], "NET") || !strcmp(av[0], "HOST") || !strcmp(av[0], "PERSON")) {
			sectype = strdup(av[0]);
			asprintf(&target, "%s/%s", av[1], av[2]);
		} else {
			sectype= NULL;
			asprintf(&target, "%s/%s/%s", av[0], av[1], av[2]);
		}
	}

	find_mode = SEARCH_SUBSTR;
	print_mode = 0;
	relation = FALSE;
	switch(target[0]) {
	case '!':
		find_mode = SEARCH_EXACT;
		strcpy(target, &target[1]);
		break;
	case '*':
		print_mode = PRINT_LONGFORM|PRINT_RELATION|PRINT_MODIFIED;
		strcpy(target, &target[1]);
		break;
	case '^':
		find_mode = SEARCH_EXACT;
		relation = TRUE;
		strcpy(target, &target[1]);
		break;
	}
	if(find_mode == SEARCH_SUBSTR || relation) {
		int len;

		len = strlen(target);
		if(len>3 && !strncmp(target, "...", 3)) {
			find_mode = SEARCH_LASTSTR;
			strcpy(target, &target[3]);
		} else if(len>3 && !strcmp(&target[len-3], "...")) {
			find_mode = SEARCH_FIRSTSTR;
			target[len-3] = '\0';
		}
	}

	ca_list = edb_catalogue_new();
	if(!strncmp(target, "EID=", 4) && str_isdigit(&target[4])) {
		EdbTupleInfo	tuple;

		tuple = edb_tuple_get(ec, ATOI(&target[4]), 0);
		if(tuple) edb_catalogue_append(ca_list, TupleEID(tuple), tuple);
	} else if(edb_text_isIPv4CIDR(ec, (utf8_t *)target)) {
		EdbTupleInfo	tuple;

		tuple = edb_dns_lookup_network_by_cidr(ec, target);
		if(tuple) edb_catalogue_append(ca_list, TupleEID(tuple), tuple);
	} else if(find_mode==SEARCH_EXACT) {
		EdbTupleInfo	tuple;

		if(sectype==NULL || !strcmp(sectype, "DOM")) {
			tuple = edb_dns_lookup_domain_by_fqdn(ec, target);
			if(tuple) edb_catalogue_append(ca_list, TupleEID(tuple), tuple);
		}
		if(sectype==NULL || !strcmp(sectype, "NET")) {
			tuple = edb_dns_lookup_network_by_cidr(ec, target);
			if(tuple) edb_catalogue_append(ca_list, TupleEID(tuple), tuple);
		}
		if(sectype==NULL || !strcmp(sectype, "HOST")) {
			tuple = edb_dns_lookup_host_by_fqdn(ec, target);
			if(tuple) edb_catalogue_append(ca_list, TupleEID(tuple), tuple);
		}
	} else {
		EdbTupleInfo	tuple;
		Link	l, l_tuple;

		if(sectype==NULL || !strcmp(sectype, "DOM")) {
			l_tuple = edb_dns_lookup_domain_by_name(ec, target, find_mode);
			ForLink(l, l_tuple) {
				if((tuple=l->d)!=NULL) edb_catalogue_append(ca_list, TupleEID(tuple), tuple);
			}
			link_free(l_tuple);
		}
		if(sectype==NULL || !strcmp(sectype, "HOST")) {
			l_tuple = edb_dns_lookup_host_by_name(ec, target, find_mode);
			ForLink(l, l_tuple) {
				if((tuple=l->d)!=NULL) edb_catalogue_append(ca_list, TupleEID(tuple), tuple);
			}
			link_free(l_tuple);

			l_tuple = edb_dns_lookup_host_by_address(ec, target, find_mode);
			ForLink(l, l_tuple) {
				if((tuple=l->d)!=NULL) edb_catalogue_append(ca_list, TupleEID(tuple), tuple);
			}
			link_free(l_tuple);
		}
	}

	edb_setdatabasespeaking(ec, japanese ? "Japanese" : "English");
	edb_caption_omit_append(ec, ORGAN_TOP_EID);
	if(edb_catalogue_size(ca_list)>0) {
		Link	l;

		doc = doc_fpopen(DocML_PLAIN, stdout);
		if(edb_catalogue_size(ca_list) == 1) {
			print_mode |= PRINT_LONGFORM|PRINT_RELATION|PRINT_MODIFIED;
		}
		ForCatalogueLink(l, ca_list) {
			EdbTupleInfo	tuple = CatalogueObject(l);

			edb_wdb_print(ec, tuple, japanese, print_mode, doc);
			if(print_mode != 0) printf("\n");
		}
		doc_fpclose(doc);
	} else {
		printf("No match for '%s'.\n", target);
	}
	edb_catalogue_free(ca_list);

	FREE(sectype);
	FREE(target);
	return TRUE;
}

static
void
html_HEAD(EdbContext ec, char *title, Doc doc)
{
	char	*bgc = "white";

	doc_puts(doc, "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n");
	doc_puts(doc, "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n");
	doc_puts(doc, "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"ja\" lang=\"ja\">\n");
	doc_puts(doc, "	<head>\n");
	doc_puts(doc, "		<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />\n");
	doc_puts(doc, "		<meta http-equiv=\"Content-Style-Type\" content=\"text/css; charset=utf-8\" />\n");
	doc_puts(doc, "		<meta http-equiv=\"Pragma\" content=\"no-cache\" />\n");
	doc_puts(doc, "		<meta http-equiv=\"Cache-Control\" content=\"no-cache\" />\n");
	doc_puts(doc, "		<meta http-equiv=\"Expires\" content=\"Thu, 01 Dec 1994 16:00:00 GMT\" />\n");
	doc_printf(doc, "		<title>%s</title>\n", title);
	doc_puts(doc, "	</head>\n");
	if(ec->university) {
		bgc = "#ccffff";	/* skyblue (light) */
	} else {
		bgc = "#80ffff";	/* skyblue (dark) */
	}
	doc_printf(doc, "<body text=\"black\" bgcolor=\"%s\" style=\"font-size:10pt\">\n", bgc);
	doc_text_puts3(doc, "<h2>", title, "</h2>\n");
	doc_puts(doc, "<hr />\n");
}

static
void
html_FOOT(EdbContext ec, Doc doc)
{
	char	buf[BUFSIZ];

	doc_puts(doc, "<hr />\n");
	ctime_r(&ec->now, buf);
	doc_puts(doc, "<span style=\"font-size:smaller\">");
	doc_printf(doc, "<i>%s &lt;%s&gt; at %s.</i>\n", EDB_WG_NAME, EDB_ADMIN_MAILADDR, buf);
	doc_printf(doc, "(%s:%s:%s, o=%d/%d, c=%d, q=%d, h=%d)\n", ec->host, ec->port, ec->dbName, ec->open_trials, link_atoms(ec->l_dbpath), ec->number_of_command, ec->number_of_query, ec->number_of_list_hit);
	doc_puts(doc, "</span>");
	doc_puts(doc, "</body>\n");
	doc_puts(doc, "</html>\n");
}

#define	LINK_SELF	0x00000001
#define	HCENTER		0x00000008

static
void
edb_dns_print_column(EdbContext ec, EdbTupleInfo tuple, char *cxn, int mode, Doc doc)
{
	EdbTupleColumn	tc;
	int	printed = FALSE;
	EdbDatum	datum;

	doc_puts(doc, "<td");
	if(mode&HCENTER) doc_puts(doc, " align=\"center\"");
	doc_puts(doc, ">");
	tc = edb_tuple_seek_column(ec, tuple, (utf8_t *)cxn);
	ForTCDatum(datum, tc) {
		if(!edb_datum_at_present(ec, datum)) continue;
		if(DatumEIDisVALID(datum)) {
			EdbTupleInfo	rtuple = edb_tuple_get(ec, DatumEID(datum), 0);
			Caption	cap;

			if(printed) doc_puts(doc, "<br />");
			if(TupleIsInetHost(rtuple)) {
				edb_dns_link_to_host(ec, rtuple, OMIT_MY_DOMAIN, doc);
			} else if(TupleIsInetDomain(rtuple)) {
				edb_dns_link_to_domain(ec, rtuple, OMIT_MY_DOMAIN, doc);
			} else if(TupleIsInetNetwork(rtuple)) {
				edb_dns_link_to_network(ec, rtuple, 0, doc);
			} else if((cap = edb_tuple_make_caption(ec, rtuple, 0))!=NULL) {
				doc_printf(doc, "<a href=\"%s?EID=%d\">", ec->browser_url, DatumEID(datum));
				doc_text_puts(doc, (char *)cap->c_main);
				edb_caption_free(ec, cap);
				doc_puts(doc, "</a>");
			} else {
				doc_printf(doc, "<a href=\"%s?EID=%d\">", ec->browser_url, DatumEID(datum));
				doc_text_printf(doc, "[%d]", DatumEID(datum));
				doc_puts(doc, "</a>");
			}
			printed = TRUE;
		} else if(mode&LINK_SELF) {
			Caption	cap;

			if(printed) doc_puts(doc, "<br />");
			if(TupleIsInetHost(tuple)) {
				edb_dns_link_to_host(ec, tuple, OMIT_MY_DOMAIN, doc);
			} else if(TupleIsInetDomain(tuple)) {
				edb_dns_link_to_domain(ec, tuple, OMIT_MY_DOMAIN, doc);
			} else if(TupleIsInetNetwork(tuple)) {
				edb_dns_link_to_network(ec, tuple, 0, doc);
			} else if((cap = edb_tuple_make_caption(ec, tuple, 0))!=NULL) {
				doc_printf(doc, "<a href=\"%s?EID=%d\">", ec->browser_url, TupleEID(tuple));
				doc_text_puts(doc, (char *)cap->c_main);
				edb_caption_free(ec, cap);
				doc_puts(doc, "</a>");
			} else {
				doc_printf(doc, "<a href=\"%s?EID=%d\">", ec->browser_url, TupleEID(tuple));
				doc_text_printf(doc, "[%d]", TupleEID(tuple));
				doc_puts(doc, "</a>");
			}
			printed = TRUE;
		} else if(DatumJapaneseIsValid(datum)) {
			if(printed) doc_puts(doc, "<br />");
			doc_text_puts(doc, (char *)DatumJapanese(datum));
			printed = TRUE;
		} else if(DatumEnglishIsValid(datum)) {
			if(printed) doc_puts(doc, "<br />");
			doc_text_puts(doc, (char *)DatumEnglish(datum));
			printed = TRUE;
		}
	}
	doc_puts(doc, "</td>");
}

static
void *
dominfo_for_compare(void *obj, void *data)
{
	EdbContext ec = data;
	EdbTupleInfo	domain = obj;
	char	*fqdn;
	Link	lstr;

	fqdn = edb_dns_domain_fqdn(ec, domain);
	lstr = str_split(fqdn, '.');
	FREE(fqdn);
	lstr = link_reverse(lstr);
	fqdn = str_concatenate(lstr, '.');
	link_free(lstr);
	return fqdn;
}

static
int
compare_domain_name(LinkObj o1, LinkObj o2)
{
	char	*n1 = o1->aux;
	char	*n2 = o2->aux;

	return strcasecmp(n1, n2);
}

static
int
edb_dns_domain_summary(EdbContext ec, Doc doc)
{
	Link	l, l_domain, l_eid;
	EdbTableInfo	ti;
	char	*query, *condition;
	int	n, hosts;
#if defined(DNS_CHECK)
	EdbCatalogue	ca_damaged_ns;
#endif

	html_HEAD(ec, "ドメイン一覧", doc);

	edb_caption_omit_append(ec, ORGAN_TOP_EID);
	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetDomain_XN);
	if(ti==NULL) return FALSE;
	asprintf(&query, "SELECT " EDB_BASE_EID " FROM %s WHERE %s",
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionRegexTextEQ("@.upper", MY_DOMAIN),
			ConditionTerminate));
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_domain = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	domain = edb_tuple_get(ec, ATOI(l->d), 0);

		if(domain) LinkAppend(l_domain, link_atom_new(domain, NULL));
	}
	link_free(l_eid);

	l_domain = link_sort(l_domain, compare_domain_name, dominfo_for_compare, ec, free);

#if defined(DNS_CHECK)
	dns_initialize();
	ca_damaged_ns = edb_catalogue_new();
#endif

	doc_puts(doc, "<table border=\"1\" width=\"100%\" style=\"font-size:smaller\">\n");
	doc_puts(doc, "<tr>");
	doc_text_puts3(doc, "<th>", "Name", "</th>");
	doc_text_puts3(doc, "<th>", "Organization", "</th>");
	doc_text_puts3(doc, "<th>", "Administrator", "</th>");
	doc_text_puts3(doc, "<th>", "Contact", "</th>");
	doc_text_puts3(doc, "<th>", "NameServer", "</th>");
	doc_text_puts3(doc, "<th>", "MailExchanger", "</th>");
	doc_text_puts3(doc, "<th>", "IP-network", "</th>");
	doc_text_puts3(doc, "<th>", "#hosts", "</th>");
	doc_puts(doc, "</tr>\n");
	hosts = 0;
	ForLink(l, l_domain) {
		EdbTupleInfo	domain = l->d;
		char	*s, *fqdn, *dom_fqdn;
		Link	l_hosts;
		EdbTupleColumn	tc;
		EdbDatum	ns;
		int	printed, gotten_ns_num;

		doc_puts(doc, "<tr>");
		doc_printf(doc, "<td>");
		edb_dns_link_to_domain(ec, domain, OMIT_MY_DOMAIN, doc);
		doc_puts(doc, "</td>");
		dom_fqdn = edb_dns_domain_fqdn(ec, domain);
		edb_dns_print_column(ec, domain, "@.organization", 0, doc);
		edb_dns_print_column(ec, domain, "@.administrator", HCENTER, doc);
		edb_dns_print_column(ec, domain, "@.contact", HCENTER, doc);
#if !defined(DNS_CHECK)
		edb_dns_print_column(ec, domain, "@.ns", 0, doc);
#else
		doc_puts(doc, "<td>");
		printed = FALSE;
		tc = edb_tuple_seek_column(ec, domain, (utf8_t *)"@.ns");
		gotten_ns_num = 0;
		ForTCDatum(ns, tc) {
			EdbTupleInfo	rtuple;
			Caption	cap;
			char	*s, *ns_fqdn, *cidr;
			int	status;
			struct	DNS_DOMAIN_INFO	dominfo;

			if(!edb_datum_at_present(ec, ns)) continue;
			if(!DatumEIDisVALID(ns)) continue;

			rtuple = edb_tuple_get(ec, DatumEID(ns), 0);
			if(!TupleIsInetHost(rtuple)) continue;
			if(printed) doc_puts(doc, "<br />");
			ns_fqdn = edb_dns_host_fqdn(ec, rtuple);
			str_minuscule1(ns_fqdn);
			if(edb_catalogue_lookup(ca_damaged_ns, TupleEID(rtuple))) {
				status = DNS_NOT_AUTHORITATIVE-1;
			} else {
				status = dns_query_domain_information(dom_fqdn, ns_fqdn, &dominfo, stderr);
			}
			if(status>0) {
				doc_puts(doc, "◎");
				++gotten_ns_num;
			} else if(status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, "<span style=\"color:#800000\">△</span>");
			} else {
				doc_puts(doc, "<span style=\"color:red\">×</span>");
				edb_catalogue_append(ca_damaged_ns, TupleEID(rtuple), rtuple);
			}
			edb_dns_link_to_host(ec, rtuple, OMIT_MY_DOMAIN, doc);
			FREE(ns_fqdn);
			printed = TRUE;
			doc_flush(doc);
		}
		doc_puts(doc, "</td>");
#endif
		edb_dns_print_column(ec, domain, "@.mx", 0, doc);
		edb_dns_print_column(ec, domain, "@.ipnetwork", 0, doc);
		l_hosts = edb_dns_zone_get_hosts(ec, domain);
		doc_printf(doc, "<td align=\"center\">%d</td>", link_atoms(l_hosts));
		hosts += link_atoms(l_hosts);
		link_free2(ec, l_hosts);
		doc_puts(doc, "</tr>\n");
	}
#if defined(DNS_CHECK)
	edb_catalogue_free(ca_damaged_ns);
#endif
	doc_puts(doc, "<tr>");
	doc_text_puts3(doc, "<th>", "Name", "</th>");
	doc_text_puts3(doc, "<th>", "Organization", "</th>");
	doc_text_puts3(doc, "<th>", "Administrator", "</th>");
	doc_text_puts3(doc, "<th>", "Contact", "</th>");
	doc_text_puts3(doc, "<th>", "NameServer", "</th>");
	doc_text_puts3(doc, "<th>", "MailExchanger", "</th>");
	doc_text_puts3(doc, "<th>", "IP-network", "</th>");
	doc_text_puts3(doc, "<th>", "#hosts", "</th>");
	doc_puts(doc, "</tr>\n");
	doc_puts(doc, "</table>\n");
	link_free(l_domain);

	html_FOOT(ec, doc);
	return TRUE;
}

struct NETINFO {
	in_addr_t 	addr;
	int	mask;
};

static
void *
netinfo_for_compare(void *obj, void *data)
{
	EdbContext ec = data;
	EdbTupleInfo	network = obj;
	char	*cidr;
	struct NETINFO *ni;

	cidr = edb_dns_network_cidr(ec, network);
	ni = (struct NETINFO *)malloc(sizeof(struct NETINFO));
	ni->addr = cidr_to_address(cidr);
	ni->mask = cidr_to_mask(cidr);
	FREE(cidr);
	return ni;
}

static
int
compare_network_cidr(LinkObj o1, LinkObj o2)
{
	struct NETINFO	*ri1 = o1->aux;
	struct NETINFO	*ri2 = o2->aux;

	if(ri1->addr<ri2->addr) return -1;
	if(ri1->addr>ri2->addr) return 1;
	if(ri1->mask<ri2->mask) return -1;
	if(ri1->mask>ri2->mask) return 1;
	return 0;
}

static
int
edb_dns_network_summary(EdbContext ec, Doc doc)
{
	Link	l, l_network, l_eid;
	EdbTableInfo	ti;
	char	*query;
	int	n;
#if defined(DNS_CHECK)
	EdbCatalogue	ca_damaged_ns;
#endif

	html_HEAD(ec, "ネットワーク一覧", doc);

	edb_caption_omit_append(ec, ORGAN_TOP_EID);
	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetNetwork_XN);
	if(ti==NULL) return FALSE;
	asprintf(&query, "SELECT " EDB_BASE_EID " FROM %s", TABLE_SN(ti));
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_network = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	network = edb_tuple_get(ec, ATOI(l->d), 0);

		if(network) LinkAppend(l_network, link_atom_new(network, NULL));
	}
	link_free(l_eid);

	l_network = link_sort(l_network, compare_network_cidr, netinfo_for_compare, ec, free);

#if defined(DNS_CHECK)
	dns_initialize();
	ca_damaged_ns = edb_catalogue_new();
#endif

	doc_puts(doc, "<table border=\"1\" width=\"100%\" style=\"font-size:smaller\">\n");
	doc_puts(doc, "<tr>");
	doc_text_puts3(doc, "<th>", "Address", "</th>");
	doc_text_puts3(doc, "<th>", "Organization", "</th>");
	doc_text_puts3(doc, "<th>", "Name", "</th>");
	doc_text_puts3(doc, "<th>", "Administrator", "</th>");
	doc_text_puts3(doc, "<th>", "Contact", "</th>");
	doc_text_puts3(doc, "<th>", "NameServer", "</th>");
	doc_text_puts3(doc, "<th>", "#hosts", "</th>");
	doc_puts(doc, "</tr>\n");
	ForLink(l, l_network) {
		EdbTupleInfo	network = l->d;
		Link	l_hosts;
		EdbTupleColumn	tc;
		EdbDatum	ns;
		int	printed, gotten_ns_num;
		in_addr_t	address;
		char		*cidr, *fqdn = NULL;
		int		mask;

		doc_puts(doc, "<tr>");
		edb_dns_print_column(ec, network, "@.address", LINK_SELF, doc);
		edb_dns_print_column(ec, network, "@.organization", 0, doc);
		edb_dns_print_column(ec, network, "@.name", 0, doc);
		edb_dns_print_column(ec, network, "@.administrator", 0, doc);
		edb_dns_print_column(ec, network, "@.contact", 0, doc);
#if !defined(DNS_CHECK)
		edb_dns_print_column(ec, network, "@.ns", 0, doc);
#else
		doc_puts(doc, "<td>");
		printed = FALSE;
		tc = edb_tuple_seek_column(ec, network, (utf8_t *)"@.ns");
		gotten_ns_num = 0;
		cidr = edb_dns_network_cidr(ec, network);
		address = cidr_to_address(cidr);
		mask = cidr_to_mask(cidr);
		if(mask<=8) {
			asprintf(&fqdn, "%d.in-addr.arpa", (address>>24)&0xff);
		} else if(mask<=16) {
			asprintf(&fqdn, "%d.%d.in-addr.arpa", (address>>16)&0xff, (address>>24)&0xff);
		} else if(mask<=24) {
			asprintf(&fqdn, "%d.%d.%d.in-addr.arpa", (address>>8)&0xff, (address>>16)&0xff, (address>>24)&0xff);
		}
		if(STRisValid(fqdn)) {
			ForTCDatum(ns, tc) {
				EdbTupleInfo	rtuple;
				Caption	cap;
				char	*s, *ns_fqdn;
				int	status;
				struct	DNS_DOMAIN_INFO	dominfo;

				if(!edb_datum_at_present(ec, ns)) continue;
				if(!DatumEIDisVALID(ns)) continue;

				rtuple = edb_tuple_get(ec, DatumEID(ns), 0);
				if(!TupleIsInetHost(rtuple)) continue;
				if(printed) doc_puts(doc, "<br />");
				ns_fqdn = edb_dns_host_fqdn(ec, rtuple);
				str_minuscule1(ns_fqdn);
				if(edb_catalogue_lookup(ca_damaged_ns, TupleEID(rtuple))) {
					status = DNS_NOT_AUTHORITATIVE-1;
				} else {
					status = dns_query_domain_information(fqdn, ns_fqdn, &dominfo, stderr);
				}
				if(status>0) {
					doc_puts(doc, "◎");
					++gotten_ns_num;
				} else if(status==DNS_NOT_AUTHORITATIVE) {
					doc_puts(doc, "<span style=\"color:#800000\">△</span>");
				} else {
					doc_puts(doc, "<span style=\"color:red\">×</span>");
					edb_catalogue_append(ca_damaged_ns, TupleEID(rtuple), rtuple);
				}
				edb_dns_link_to_host(ec, rtuple, OMIT_MY_DOMAIN, doc);
				FREE(ns_fqdn);
				printed = TRUE;
				doc_flush(doc);
			}
		}
		doc_puts(doc, "</td>");
		FREE(fqdn);
#endif
		l_hosts = edb_dns_lookup_hosts_by_network(ec, network);
		doc_printf(doc, "<td align=\"center\">%d</td>", link_atoms(l_hosts));
		link_free2(ec, l_hosts);
		doc_puts(doc, "</tr>\n");
	}
#if defined(DNS_CHECK)
	edb_catalogue_free(ca_damaged_ns);
#endif
	doc_puts(doc, "<tr>");
	doc_text_puts3(doc, "<th>", "Address", "</th>");
	doc_text_puts3(doc, "<th>", "Organization", "</th>");
	doc_text_puts3(doc, "<th>", "Name", "</th>");
	doc_text_puts3(doc, "<th>", "Administrator", "</th>");
	doc_text_puts3(doc, "<th>", "Contact", "</th>");
	doc_text_puts3(doc, "<th>", "NameServer", "</th>");
	doc_text_puts3(doc, "<th>", "#hosts", "</th>");
	doc_puts(doc, "</tr>\n");
	doc_puts(doc, "</table>\n");
	link_free(l_network);

	html_FOOT(ec, doc);
	return TRUE;
}

#define	NS_TYPE_NONE		0
#define	NS_TYPE_PRIMARY		1
#define	NS_TYPE_SECONDARY	2

static
int
edb_dns_make_nsconf(EdbContext ec, char *nameserver, FILE *fp)
{
	EdbTableInfo	ti_domain, ti_network;
	EdbTupleInfo	ns;
	EdbColumnInfo	ci;
	char	*condition, *order;
	Link	l, l_eid, l_domain, l_network;

	ns = edb_dns_lookup_host_by_fqdn(ec, nameserver);
	if(ns==NULL) {
		fprintf(fp, "; can't get information of nameserver (%s).\n", nameserver);
		return FALSE;
	}

	ti_domain = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetDomain_XN);
	if(ti_domain==NULL) {
		fprintf(fp, "; Internal Error : table information for domain load failed.\n");
		return FALSE;
	}
	ci = edb_table_seek_column_by_xmlname(ec, ti_domain, (utf8_t *)"@.name");
	condition = (char *)edb_condition_make(ec, ti_domain, CONDITION_LOGIC_AND,
		ConditionAvailable,
#if 0
		ConditionEIDMatch("@.ns", TupleEID(ns)),
#endif
		ConditionTerminate);
	l_domain = edb_dns_pgsql_query(ec, ti_domain, condition, (char *)ColumnSN(ci));
	FREE(condition);
	l_domain = link_sort(l_domain, compare_domain_name, dominfo_for_compare, ec, free);
	ForLink(l, l_domain) {
		EdbTupleInfo	domain = l->d;
		EdbTupleColumn	tc;
		EdbDatum	datum;
		char	*fqdn;
		int	ns_type = NS_TYPE_NONE;
		EdbCatalogue	ca_ns = edb_catalogue_new();

		tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.ns");
		ForTCDatum(datum, tc) {
			EdbCatalogue	ca;
			Link	ln;

			ca = edb_dns_lookup_iphosts_by_EID(ec, DatumEID(datum), ec->max_mapto_recurse);
			ForCatalogueLink(ln, ca) {
				EdbTupleInfo	tuple = CatalogueObject(ln);

				if(tuple) edb_catalogue_append(ca_ns, TupleEID(tuple), tuple);
			}
			if(ca && edb_catalogue_lookup(ca, TupleEID(ns))) {
				ns_type = TCDatumIsFirst(tc, datum) ? NS_TYPE_PRIMARY : NS_TYPE_SECONDARY;
				edb_catalogue_free(ca);
				break;
			}
			edb_catalogue_free(ca);
		}
		if(ns_type==NS_TYPE_NONE) continue;

		fqdn = edb_dns_domain_fqdn(ec, domain);
		if(!STRisVALID(fqdn)) {
			FREE(fqdn);
			continue;
		}
		str_minuscule1(fqdn);
		fprintf(fp, "%s/ZONE/%s",
			(ns_type==NS_TYPE_PRIMARY) ? "PRIMARY" : "SECONDARY",
			fqdn);
		if(ns_type==NS_TYPE_SECONDARY) {
#if 1
			Link	ln;

			ForCatalogueLink(ln, ca_ns) {
				EdbTupleInfo	master = CatalogueObject(ln);
				char	*address;

				if(master==NULL) continue;
				if(TupleEID(master)==TupleEID(ns)) continue;
				address = edb_dns_host_address(ec, master);
				if(STRisVALID(address)) fprintf(fp, "/%s", address);
				FREE(address);
			}
#else
			ForTCDatum(datum, tc) {
				EdbTupleInfo	master;
				char	*address;

				if(DatumEID(datum)==TupleEID(ns)) continue;
				if(!DatumEIDisVALID(datum)) continue;
				master = edb_tuple_get(ec, DatumEID(datum), 0);
				if(master==NULL) continue;
				address = edb_dns_host_address(ec, master);
				if(STRisVALID(address)) fprintf(fp, "/%s", address);
				FREE(address);
			}
#endif
		}
		fprintf(fp, "\n");
		FREE(fqdn);
		edb_catalogue_free(ca_ns);
	}
	link_free(l_domain);

	ti_network = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetNetwork_XN);
	if(ti_network==NULL) {
		fprintf(fp, "; Internal Error : table information for network load failed.\n");
		return FALSE;
	}
	condition =  (char *)edb_condition_make(ec, ti_network, CONDITION_LOGIC_AND,
#if 0
		ConditionEIDMatch("@.ns", TupleEID(ns)),
#endif
		ConditionTerminate);
	ci = edb_table_seek_column_by_xmlname(ec, ti_network, (utf8_t *)"@.address");
	asprintf(&order, "%s::inet", ColumnSN(ci));
	l_network = edb_dns_pgsql_query(ec, ti_network, condition, order);
	FREE(condition);
	FREE(order);
	l_network = link_sort(l_network, compare_network_cidr, netinfo_for_compare, ec, free);
	ForLink(l, l_network) {
		EdbTupleInfo	network = l->d;
		EdbTupleColumn	tc;
		EdbDatum	datum;
		char	*cidr;
		int	ns_type = NS_TYPE_NONE;
		EdbCatalogue	ca_ns = edb_catalogue_new();

		tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.ns");
		ForTCDatum(datum, tc) {
			EdbCatalogue	ca;
			Link	ln;

			ca = edb_dns_lookup_iphosts_by_EID(ec, DatumEID(datum), ec->max_mapto_recurse);
			ForCatalogueLink(ln, ca) {
				EdbTupleInfo	tuple = CatalogueObject(ln);

				if(tuple) edb_catalogue_append(ca_ns, TupleEID(tuple), tuple);
			}
			if(ca && edb_catalogue_lookup(ca, TupleEID(ns))) {
				ns_type = TCDatumIsFirst(tc, datum) ? NS_TYPE_PRIMARY : NS_TYPE_SECONDARY;
				edb_catalogue_free(ca);
				break;
			}
			edb_catalogue_free(ca);
		}
		if(ns_type==NS_TYPE_NONE) continue;

		cidr = edb_dns_network_cidr(ec, network);
		if(!STRisVALID(cidr)) {
			FREE(cidr);
			continue;
		}
		str_minuscule1(cidr);
		fprintf(fp, "%s/REV/%s",
			(ns_type==NS_TYPE_PRIMARY) ? "PRIMARY" : "SECONDARY",
			cidr);
		if(ns_type==NS_TYPE_SECONDARY) {
#if 1
			Link	ln;

			ForCatalogueLink(ln, ca_ns) {
				EdbTupleInfo	master = CatalogueObject(ln);
				char	*address;

				if(master==NULL) continue;
				if(TupleEID(master)==TupleEID(ns)) continue;
				address = edb_dns_host_address(ec, master);
				if(STRisVALID(address)) fprintf(fp, "/%s", address);
				FREE(address);
			}
#else
			ForTCDatum(datum, tc) {
				EdbTupleInfo	master;
				char	*address;

				if(DatumEID(datum)==TupleEID(ns)) continue;
				if(!DatumEIDisVALID(datum)) continue;
				master = edb_tuple_get(ec, DatumEID(datum), 0);
				if(master==NULL) continue;
				address = edb_dns_host_address(ec, master);
				if(STRisVALID(address)) fprintf(fp, "/%s", address);
				FREE(address);
			}
#endif
		}
		fprintf(fp, "\n");
		FREE(cidr);
		edb_catalogue_free(ca_ns);
	}
	link_free(l_network);

	fprintf(fp, ";\n");
	fprintf(fp, "%s\n", END_OF_NSCONF);
	fprintf(fp, ";\n");
	return TRUE;
}

static
int
edb_dns_script_print(EdbContext ec, char *scriptname, FILE *fp)
{
	struct SCRIPT2FILE {
		char	*name;
		char	*file;
	} script2file [] = {
	{	"edb-dns-conf.sh",	EDBDIR "/libexec/dns/edb-dns-conf.sh"	},
	{	"dns-update.sh",	EDBDIR "/libexec/dns/dns-update.sh"	},
	};
	int	i;

	for(i=0; i<Ns(script2file); ++i) {
		if(!strcmp(scriptname, script2file[i].name)) {
			Link	l, lstr;

			lstr = file_load_as_linked_str(script2file[i].file, NULL);
			ForLink(l, lstr) fprintf(fp, "%s\n", l->d);
			link_free(lstr);
			return TRUE;
		}
	}
	return FALSE;
}

static
int
edb_dns_domain_info(EdbContext ec, char *fqdn, Doc doc)
{
	char	*title;
	EdbTupleInfo	domain;
	EdbDatum	datum = NULL;
	EdbTupleColumn	tc;
	EdbCatalogue	ca_mx_printed;
	int	preference;
	char	*master_ip;
	Link	l, l_domains, l_hosts;

	asprintf(&title, "Domain / %s のDNS関連情報", fqdn);
	html_HEAD(ec, title, doc);
	FREE(title);
	doc_puts(doc, "<blockquote>\n");

	domain = edb_dns_lookup_domain_by_fqdn(ec, fqdn);
	if(domain==NULL) {
		doc_printf(doc, "<span style=\"color:red\">ドメイン(%s)の情報が見つかりません．</span>\n", fqdn);
		doc_puts(doc, "</blockquote>\n");
		html_FOOT(ec, doc);
		return FALSE;
	}

#if defined(DNS_CHECK)
	dns_initialize();
#endif

	doc_puts(doc, "<dl>");
	doc_printf(doc, "<dt>◇ 登録情報(EDB/DNS): <a href=\"%s?EID=%d\">%s</a></dt>\n", ec->browser_url, TupleEID(domain), fqdn);

	doc_printf(doc, "<dt>◇ DNSサーバ(BIND)におけるこのドメインの正引き設定: <a href=\"/DNS/BIND/ZONE/%s\">/DNS/BIND/ZONE/%s</a></dt>\n", fqdn, fqdn);

	doc_printf(doc, "<dt>◇ 所属ドメイン: ");
	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)DOMAIN_C_UPPER);
	ForTCDatum(datum, tc) {
		EdbTupleInfo	dom = edb_tuple_get(ec, DatumEID(datum), 0);

		if(dom==NULL) continue;
		edb_dns_link_to_domain(ec, dom, 0, doc);
	}
	doc_puts(doc, "</dt>\n");

	doc_printf(doc, "<dt>◇ このドメイン宛ての電子メールについて</dt>\n");
	doc_printf(doc, "<dd>DNSのMX (Mail eXchanger)情報を参照するMail Delivery Agentは，");
	doc_printf(doc, "このドメイン宛ての電子メール(宛先: <var style=\"color:green\">username</var>@<u>%s</u> )を中継する場合，", fqdn);
	doc_printf(doc, "DNSより次段の電子メール中継サーバの情報として\n", fqdn);

	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.mx");
	if(tc) datum = tc->datum;
	doc_puts(doc, "<ul>\n");
	preference = 10;
	ca_mx_printed = edb_catalogue_new();
	ForTCDatum(datum, tc) {
		EdbTupleInfo	host;
		EdbCatalogue	ca_mx;

		if(!DatumEIDisVALID(datum)) continue;
		ca_mx = edb_dns_lookup_iphosts_by_EID(ec, DatumEID(datum), ec->max_mapto_recurse);
		if(ca_mx) {
			Link	lm;

			ForCatalogueLink(lm, ca_mx) {
				host = CatalogueObject(lm);
				if(edb_catalogue_lookup(ca_mx_printed, TupleEID(host))) continue;
				doc_puts(doc, "<li>");
				edb_dns_link_to_host(ec, host, 0, doc);
				doc_printf(doc, "&nbsp;&nbsp;(優先度: %d)\n", preference);
				edb_catalogue_append(ca_mx_printed, TupleEID(host), host);
			}
			edb_catalogue_free(ca_mx);
		}
		++preference;
	}
	edb_catalogue_free(ca_mx_printed);
	doc_puts(doc, "</ul>\n");
	doc_printf(doc, "を取得し，これらのサーバのうち優先度が一番高い(数値では小さい)サーバにメールを送信します．");
	doc_printf(doc, "ただし，メール送信に失敗した場合には，次に優先度の高いサーバについて送信を試み，送信できなければこの作業を順次繰り返します．\n");
	doc_printf(doc, "なお，Mail Delivery Agentがこれらのサーバ自身の場合には自身の優先度よりも高いサーバにのみ送信を試みます．\n");
	doc_printf(doc, "また，Mail Delivery Agent自身が一番優先度の高いサーバである場合には，そのMail Delivery Agentの設定によります．\n");
	doc_puts(doc, "</dd>\n");

	doc_printf(doc, "<dt>◇ このドメインは下記のDNSサーバ(ネームサーバ)によってDNS情報が提供されています．</dt>\n");
	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.ns");
	master_ip = NULL;
	ForTCDatum(datum, tc) {
		struct	DNS_DOMAIN_INFO	dominfo;
		EdbCatalogue	ca_ns;
		int	primary = TCDatumIsFirst(tc, datum);
		Link	l;

		ca_ns = edb_dns_lookup_iphosts_by_EID(ec, DatumEID(datum), ec->max_mapto_recurse);
		if(primary) {
			if(edb_catalogue_size(ca_ns)>0) {
				master_ip = edb_dns_host_address(ec, (EdbTupleInfo)ca_ns->link->d);
			} else {
				master_ip = strdup("XX.XX.XX.XX");
			}
		}
		ForCatalogueLink(l, ca_ns) {
			EdbTupleInfo	host = CatalogueObject(l);
			int	status;
			char	*host_fqdn;

			doc_puts(doc, "<dd>");

			host_fqdn = edb_dns_host_fqdn(ec, host);
			str_minuscule1(host_fqdn);
			status = dns_query_domain_information(fqdn, host_fqdn, &dominfo, stderr);
			if(status>0) {
				doc_puts(doc, "◎");
			} else if(status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, "<span style=\"color:#800000\">△</span>");
			} else {
				doc_puts(doc, "<span style=\"color:red\">×</span>");
			}
			doc_puts(doc, " ");
			edb_dns_link_to_host(ec, host, 0, doc);
			doc_puts(doc, primary ? " (PRIMARY)" : " (secondary)");
			if(status>0) {
			} else if(status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, " <span style=\"color:#800000\">DNSサーバ(ネームサーバ)の設定がされていないようです．</span>");
			} else {
				doc_puts(doc, " <span style=\"color:red\">DNSサーバ(ネームサーバ)が応答しません．</span>");
			}
			doc_puts(doc, "</dd>\n");
			if(!primary && status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, "<dd><dl><dd><table border=\"1\"><tr><td>\n");
				doc_puts(doc, "このゾーンに関するnamed.conf (BIND)のセカンダリ設定は，<pre>");
				doc_printf(doc, "zone \"%s\" {\n", fqdn);
				doc_puts(doc, "	type slave;\n");
				doc_printf(doc, "	file \"back/%s.zone\";\n", fqdn);
				doc_printf(doc, "	masters { %s; };\n", master_ip);
				doc_puts(doc, "};\n");
				doc_puts(doc, "</pre>のような記述になります．\n");
				doc_puts(doc, "</td></tr></table></dd></dl></dd>\n");
			}
			FREE(host_fqdn);
		}
		edb_catalogue_free(ca_ns);
		doc_flush(doc);
	}
	FREE(master_ip);

	doc_printf(doc, "<dt>◇ このドメインには下記のネットワークのホストを登録できます．</dt>\n");
	tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)DOMAIN_C_IPNETWORK);
	ForTCDatum(datum, tc) {
		EdbTupleInfo	network = edb_tuple_get(ec, DatumEID(datum), 0);

		if(network==NULL) continue;
		doc_puts(doc, "<dd>○ ");
		edb_dns_link_to_network(ec, network, 0, doc);
		doc_puts(doc, "</dd>");
	}

	l_domains = edb_dns_zone_get_subdomains(ec, domain);
	if(l_domains) {
		doc_printf(doc, "<dt>◇ このドメインには下記のサブドメインが登録されています．</dt>\n");
		ForLink(l, l_domains) {
			EdbTupleInfo	dom = l->d;

			if(dom==NULL) continue;
			doc_puts(doc, "<dd>○ ");
			edb_dns_link_to_domain(ec, dom, 0, doc);
			doc_puts(doc, "</dd>");
		}
		link_free(l_domains);
	}

	l_hosts = edb_dns_zone_get_hosts(ec, domain);
	if(l_hosts) {
		doc_printf(doc, "<dt>◇ このドメインには下記のホストが登録されています．</dt>\n");
		doc_puts(doc, "<dd><table border=\"0\">");
		doc_puts(doc, "<tr><th>Name</th><th>IP-address</th><th>Canonical-name</th></tr>\n");
		ForLink(l, l_hosts) {
			EdbTupleInfo	host = l->d;
			char	*address;

			doc_puts(doc, "<tr><td>◇ ");
			edb_dns_link_to_host(ec, host, 0, doc);
			doc_puts(doc, "</td><td>");
			address = edb_dns_host_address(ec, host);
			if(STRisValid(address)) {
				doc_printf(doc, " [ %s ] ", address);
				doc_puts(doc, "</td><td>");
			} else {
				doc_puts(doc, "</td><td>");
				tc = edb_tuple_seek_column(ec, host, (utf8_t *)HOST_C_CNAME);
				if(tc) {
					int	num = 0;

					doc_puts(doc, " → ");
					ForTCDatum(datum, tc) {
						EdbTupleInfo	chost = edb_tuple_get(ec, DatumEID(datum), 0);

						if(chost==NULL) continue;
						if(num>0) doc_puts(doc, " / ");
						edb_dns_link_to_host(ec, chost, 0, doc);
						++num;
					}
				}
			}
			FREE(address);
			doc_puts(doc, "</td></tr>\n");
		}
		link_free(l_hosts);
		doc_puts(doc, "</table></dd>");
	}

	doc_puts(doc, "</dl>");

	doc_puts(doc, "</blockquote>\n");
	html_FOOT(ec, doc);
	return TRUE;
}

static
int
edb_dns_network_info(EdbContext ec, char *cidr, Doc doc)
{
	char	*title;
	EdbTupleInfo	network;
	EdbDatum	datum = NULL;
	EdbTupleColumn	tc;
	EdbCatalogue	ca_mx_printed;
	int	preference;
	char	*master_ip;
	Link	l, l_hosts;

	asprintf(&title, "Network / %s のDNS関連情報", cidr);
	html_HEAD(ec, title, doc);
	FREE(title);
	doc_puts(doc, "<blockquote>\n");

	network = edb_dns_lookup_network_by_cidr(ec, cidr);
	if(network==NULL) {
		doc_printf(doc, "<span style=\"color:red\">ネットワーク(%s)の情報が見つかりません．</span>\n", cidr);
		doc_puts(doc, "</blockquote>\n");
		html_FOOT(ec, doc);
		return FALSE;
	}

#if defined(DNS_CHECK)
	dns_initialize();
#endif

	doc_puts(doc, "<dl>");
	doc_printf(doc, "<dt>◇ 登録情報(EDB/DNS): <a href=\"%s?EID=%d\">%s</a></dt>\n", ec->browser_url, TupleEID(network), cidr);

	doc_printf(doc, "<dt>◇ DNSサーバ(BIND)におけるこのネットワークの逆引き設定: <a href=\"/DNS/BIND/REV/%s\">/DNS/BIND/REV/%s</a></dt>\n", cidr, cidr);

	doc_printf(doc, "<dt>◇ このネットワークは下記のDNSサーバ(ネームサーバ)によってDNS情報が提供されています．</dt>\n");
	tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.ns");
	master_ip = NULL;
	ForTCDatum(datum, tc) {
		struct	DNS_DOMAIN_INFO	dominfo;
		EdbCatalogue	ca_ns;
		int	primary = TCDatumIsFirst(tc, datum);
		Link	l;

		ca_ns = edb_dns_lookup_iphosts_by_EID(ec, DatumEID(datum), ec->max_mapto_recurse);
		if(primary) {
			if(edb_catalogue_size(ca_ns)>0) {
				master_ip = edb_dns_host_address(ec, (EdbTupleInfo)ca_ns->link->d);
			} else {
				master_ip = strdup("XX.XX.XX.XX");
			}
		}
		ForCatalogueLink(l, ca_ns) {
			EdbTupleInfo	host = CatalogueObject(l);
			int	status;
			char	*host_fqdn;
			char	*net_fqdn;
			in_addr_t	address;
			int	mask;
			char	*revname;

			doc_puts(doc, "<dd>");

			host_fqdn = edb_dns_host_fqdn(ec, host);
			str_minuscule1(host_fqdn);
			doc_puts(doc, "<dd>");
			cidr = edb_dns_network_cidr(ec, network);
			address = cidr_to_address(cidr);
			mask = cidr_to_mask(cidr);
			if(mask<=8) {
				asprintf(&net_fqdn, "%d.in-addr.arpa", (address>>24)&0xff);
				asprintf(&revname, "%d.rev", (address>>24)&0xff);
			} else if(mask<=16) {
				asprintf(&net_fqdn, "%d.%d.in-addr.arpa", (address>>16)&0xff, (address>>24)&0xff);
				asprintf(&revname, "%d.%d.rev", (address>>24)&0xff, (address>>16)&0xff);
			} else if(mask<=24) {
				asprintf(&net_fqdn, "%d.%d.%d.in-addr.arpa", (address>>8)&0xff, (address>>16)&0xff, (address>>24)&0xff);
				asprintf(&revname, "%d.%d.%d.rev", (address>>24)&0xff, (address>>16)&0xff, (address>>8)&0xff);
			} else {
				continue;
			}

			status = dns_query_domain_information(net_fqdn, host_fqdn, &dominfo, stderr);
			if(status>0) {
				doc_puts(doc, "◎");
			} else if(status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, "<span style=\"color:#800000\">△</span>");
			} else {
				doc_puts(doc, "<span style=\"color:red\">×</span>");
			}
			doc_puts(doc, " ");
			edb_dns_link_to_host(ec, host, 0, doc);
			doc_puts(doc, primary ? " (PRIMARY)" : " (secondary)");
			if(status>0) {
			} else if(status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, " <span style=\"color:#800000\">DNSサーバ(ネームサーバ)の設定がされていないようです．</span>");
			} else {
				doc_puts(doc, " <span style=\"color:red\">DNSサーバ(ネームサーバ)が応答しません．</span>");
			}
			doc_puts(doc, "</dd>\n");
			if(!primary && status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, "<dd><dl><dd><table border=\"1\"><tr><td>\n");
				doc_puts(doc, "このゾーンに関するnamed.conf (BIND)のセカンダリ設定は，<pre>");
				doc_printf(doc, "zone \"%s\" {\n", net_fqdn);
				doc_puts(doc, "	type slave;\n");
				doc_printf(doc, "	file \"back/%s\";\n", revname);
				doc_printf(doc, "	masters { %s; };\n", master_ip);
				doc_puts(doc, "};\n");
				doc_puts(doc, "</pre>のような記述になります．\n");
				doc_puts(doc, "</td></tr></table></dd></dl></dd>\n");
			}
			FREE(revname);
			FREE(net_fqdn);
			FREE(host_fqdn);
		}
		edb_catalogue_free(ca_ns);
		doc_flush(doc);
	}
	FREE(master_ip);

	l_hosts = edb_dns_rev_get_hosts(ec, network);
	if(l_hosts) {
		doc_printf(doc, "<dt>◇ このネットワークには下記のホストが登録されています．</dt>\n");
		doc_puts(doc, "<dd><table border=\"0\">");
		doc_puts(doc, "<tr><th>IP-address</th><th>Name</th></tr>\n");
		l_hosts = link_sort(l_hosts, compare_revinfo, NULL, NULL, NULL);
		ForLink(l, l_hosts) {
			RevInfo	rev = l->d;

			doc_puts(doc, "<tr><td>");
			doc_printf(doc, "○ %d.%d.%d.%d ",
				(rev->addr>>24)&0xff,
				(rev->addr>>16)&0xff,
				(rev->addr>>8)&0xff,
				(rev->addr>>0)&0xff);
			doc_puts(doc, "</td><td>");
			doc_puts(doc, " ( ");
			edb_dns_link_to_host(ec, rev->host, 0, doc);
			doc_puts(doc, " )");
			doc_puts(doc, "</td></tr>");
		}
		link_free(l_hosts);
		doc_puts(doc, "</table></dd>\n");
	}

	doc_puts(doc, "</dl>");

	doc_puts(doc, "</blockquote>\n");
	html_FOOT(ec, doc);
	return TRUE;
}

static
int
edb_dns_host_info(EdbContext ec, char *fqdn, Doc doc)
{
	Link	l, l_domain, l_eid, l_ns_domain, l_network, l_ns_network, l_mx_domain, l_host;
	EdbTableInfo	ti;
	EdbTupleColumn	tc;
	EdbDatum	datum;
	char	*query, *condition;
	char	*title;
	EdbTupleInfo	host;
	int	n;
	char	*ipaddr;

	asprintf(&title, "Host / %s のDNS関連情報", fqdn);
	html_HEAD(ec, title, doc);
	FREE(title);
	doc_puts(doc, "<blockquote>\n");

	host = edb_dns_lookup_host_by_fqdn(ec, fqdn);
	if(host==NULL) {
		doc_printf(doc, "<span style=\"color:red\">ホスト(%s)の情報が見つかりません．</span>\n", fqdn);
		doc_puts(doc, "</blockquote>\n");
		html_FOOT(ec, doc);
		return FALSE;
	}

#if defined(DNS_CHECK)
	dns_initialize();
#endif

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetDomain_XN);
	if(ti==NULL) return FALSE;
	asprintf(&query, "SELECT " EDB_BASE_EID " FROM %s WHERE %s",
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable, ConditionTerminate));
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_domain = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	domain = edb_tuple_get(ec, ATOI(l->d), 0);

		if(domain) LinkAppend(l_domain, link_atom_new(domain, NULL));
	}
	link_free(l_eid);
	l_domain = link_sort(l_domain, compare_domain_name, dominfo_for_compare, ec, free);

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_InetNetwork_XN);
	if(ti==NULL) return FALSE;
	asprintf(&query, "SELECT " EDB_BASE_EID " FROM %s WHERE %s",
		TABLE_SN(ti),
		condition =  (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable, ConditionTerminate));
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_network = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	network = edb_tuple_get(ec, ATOI(l->d), 0);

		if(network) LinkAppend(l_network, link_atom_new(network, NULL));
	}
	link_free(l_eid);
	l_network = link_sort(l_network, compare_network_cidr, netinfo_for_compare, ec, free);

	doc_puts(doc, "<dl>");
	doc_printf(doc, "<dt>◇ 登録情報(EDB/DNS): <a href=\"%s?EID=%d\">%s</a></dt>\n", ec->browser_url, TupleEID(host), fqdn);

	doc_printf(doc, "<dt>◇ 所属ドメイン: ");
	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)HOST_C_DOMAIN);
	ForTCDatum(datum, tc) {
		EdbTupleInfo	dom = edb_tuple_get(ec, DatumEID(datum), 0);

		if(dom==NULL) continue;
		edb_dns_link_to_domain(ec, dom, 0, doc);
	}
	doc_puts(doc, "</dt>\n");

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)HOST_C_ADDRESS);
	ForTCDatum(datum, tc) {
		EdbTupleInfo	net;
		in_addr_t	addr;
		char	*cidr;

		if(!DatumEnglishIsValid(datum)) continue;
		doc_printf(doc, "<dt>◇ IPアドレス: %s", DatumEnglish(datum));
		addr = inet_addr((char *)DatumEnglish(datum));
		addr = ntohl(addr);
		asprintf(&cidr, "%d.%d.%d.0/24", (addr>>24)&0xff, (addr>>16)&0xff, (addr>>8)&0xff);
		net = edb_dns_lookup_network_by_cidr(ec, cidr);
		if(net) {
			doc_puts(doc, " @ ");
			edb_dns_link_to_network(ec, net, 0, doc);
		}
		FREE(cidr);
		doc_puts(doc, "</dt>\n");
	}

	tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)HOST_C_CNAME);
	ForTCDatum(datum, tc) {
		EdbTupleInfo	chost;

		if(!DatumEIDisValid(datum)) continue;
		chost = edb_tuple_get(ec, DatumEID(datum), 0);
		if(chost==NULL) continue;
		doc_puts(doc, "<dt>◇ 正式名: → ");
		edb_dns_link_to_host(ec, chost, 0, doc);
		doc_puts(doc, "</dt>\n");
	}

	ti = TupleTable(host);
	asprintf(&query, "SELECT " EDB_BASE_EID " FROM %s WHERE %s",
		TABLE_SN(ti),
		condition = (char *)edb_condition_make(ec, ti, CONDITION_LOGIC_AND,
			ConditionAvailable,
			ConditionEIDMatch(HOST_C_CNAME, TupleEID(host)),
			ConditionTerminate));
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	FREE(query);
	l_host = NULL;
	ForLink(l, l_eid) {
		EdbTupleInfo	h = edb_tuple_get(ec, ATOI(l->d), 0);

		if(h) LinkAppend(l_host, link_atom_new(h, NULL));
	}
	link_free(l_eid);
	if(l_host) {
		doc_puts(doc, "<dt>◇ このホストの別称として以下のホストが登録されています．</dt>\n");
		ForLink(l, l_host) {
			EdbTupleInfo	ahost = l->d;

			doc_puts(doc, "<dd>◇ ← ");
			edb_dns_link_to_host(ec, ahost, 0, doc);
			doc_puts(doc, "</dd>\n");
		}
	}

	ipaddr = edb_dns_host_address(ec, host);
	if(STRisValid(ipaddr)) {
		EdbDatum	datum = NULL;
		EdbTupleColumn	tc;

		doc_printf(doc, "<dt>◇ このホスト宛ての電子メールについて</dt>\n");
		doc_printf(doc, "<dd>DNSのMX (Mail eXchanger)情報を参照するMail Delivery Agentは，");
		doc_printf(doc, "このホスト宛ての電子メール(宛先: <var style=\"color:green\">username</var>@<u>%s</u> )を中継する場合，", fqdn);
		doc_printf(doc, "DNSより次段の電子メール中継サーバの情報として\n", fqdn);

		tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.mx");
		if(tc) datum = tc->datum;
		if(datum==NULL || !DatumEIDisVALID(datum)) {
			EdbTupleInfo	domain;

			tc = edb_tuple_seek_column_by_xmlname(ec, host, (utf8_t *)"@.domain");
			if(tc && tc->datum && DatumEIDisVALID(tc->datum)) {
				domain = edb_tuple_get(ec, DatumEID(tc->datum), 0);
				if(domain) {
					tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.mx");
					if(tc) datum = tc->datum;
				}
			} else {
				datum = NULL;
			}
		}
		doc_puts(doc, "<ul>\n");
		if(datum && DatumEIDisVALID(datum)) {
			EdbCatalogue	ca_mx = edb_dns_lookup_iphosts_by_EID(ec, DatumEID(datum), ec->max_mapto_recurse);

			if(ca_mx) {
				Link	lm;

				ForCatalogueLink(lm, ca_mx) {
					EdbTupleInfo	mx_host = CatalogueObject(lm);

					doc_puts(doc, "<li>");
					edb_dns_link_to_host(ec, mx_host, 0, doc);
					doc_puts(doc, "\n");
					break;
				}
				edb_catalogue_free(ca_mx);
			}
		}
		doc_puts(doc, "</ul>\n");
		doc_printf(doc, "を取得し，このサーバにメールを送信します．");
		doc_printf(doc, "ただし，Mail Delivery Agentがそのサーバ自身の場合にはそのMail Delivery Agentの設定によります．\n");
		doc_puts(doc, "</dd>\n");
	}
	doc_flush(doc);

	l_ns_domain = NULL;
	ForLink(l, l_domain) {
		EdbTupleInfo	domain = l->d;
		EdbTupleColumn	tc;
		EdbCatalogue	ca_ns;

		tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.ns");
		if(tc==NULL) continue;
		ca_ns = edb_dns_lookup_iphosts_by_TC(ec, tc);
		if(edb_catalogue_lookup(ca_ns, TupleEID(host))) LinkAppend(l_ns_domain, link_atom_new(domain, NULL));
		edb_catalogue_free(ca_ns);
	}
	if(l_ns_domain) {
		doc_printf(doc, "<dt>◇ このホスト(%s)は下記のドメインのDNSサーバ(ネームサーバ)として運用されていなければなりません．</dt>\n", fqdn);
		ForLink(l, l_ns_domain) {
			EdbTupleInfo	domain = l->d;
			char	*dom_fqdn;
			int	status;
			struct	DNS_DOMAIN_INFO	dominfo;
			EdbTupleColumn	tc;
			EdbCatalogue	ca_ns;
			int	primary = FALSE;
			char	*master_ip;

			tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.ns");
			if(tc==NULL) continue;
			ca_ns = edb_dns_lookup_iphosts_by_EID(ec, DatumEID(TCDatumFirst(tc)), ec->max_mapto_recurse);
			if(edb_catalogue_lookup(ca_ns, TupleEID(host))) primary = TRUE;
			if(edb_catalogue_size(ca_ns)>0) {
				master_ip = edb_dns_host_address(ec, (EdbTupleInfo)ca_ns->link->d);
			} else {
				master_ip = "XX.XX.XX.XX";
			}
			edb_catalogue_free(ca_ns);

			doc_puts(doc, "<dd>");
			dom_fqdn = edb_dns_domain_fqdn(ec, domain);

			status = dns_query_domain_information(dom_fqdn, fqdn, &dominfo, stderr);
			if(status>0) {
				doc_puts(doc, "◎");
			} else if(status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, "<span style=\"color:#800000\">△</span>");
			} else {
				doc_puts(doc, "<span style=\"color:red\">×</span>");
			}
			doc_puts(doc, " ");
			edb_dns_link_to_domain(ec, domain, 0, doc);
			doc_puts(doc, primary ? " (PRIMARY)" : " (secondary)");
			if(status>0) {
			} else if(status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, " <span style=\"color:#800000\">DNSサーバ(ネームサーバ)の設定がされていないようです．</span>");
			} else {
				doc_puts(doc, " <span style=\"color:red\">DNSサーバ(ネームサーバ)が応答しません．</span>");
			}
			doc_puts(doc, "</dd>\n");
			if(!primary && status==DNS_NOT_AUTHORITATIVE) {
				str_minuscule1(dom_fqdn);
				doc_puts(doc, "<dd><dl><dd><table border=\"1\"><tr><td>\n");
				doc_puts(doc, "このゾーンに関するnamed.conf (BIND)のセカンダリ設定は，<pre>");
				doc_printf(doc, "zone \"%s\" {\n", dom_fqdn);
				doc_puts(doc, "	type slave;\n");
				doc_printf(doc, "	file \"back/%s.zone\";\n", dom_fqdn);
				doc_printf(doc, "	masters { %s; };\n", master_ip);
				doc_puts(doc, "};\n");
				doc_puts(doc, "</pre>のような記述になります．\n");
				doc_puts(doc, "</td></tr></table></dd></dl></dd>\n");
			}
			FREE(dom_fqdn);
			FREE(master_ip);
			doc_flush(doc);
		}
		link_free(l_ns_domain);
	}

	l_ns_network = NULL;
	ForLink(l, l_network) {
		EdbTupleInfo	network = l->d;
		EdbTupleColumn	tc;
		EdbCatalogue	ca_ns;

		tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.ns");
		if(tc==NULL) continue;
		ca_ns = edb_dns_lookup_iphosts_by_TC(ec, tc);
		if(edb_catalogue_lookup(ca_ns, TupleEID(host))) LinkAppend(l_ns_network, link_atom_new(network, NULL));
		edb_catalogue_free(ca_ns);
	}
	if(l_ns_network) {
		doc_printf(doc, "<dt>◇ このホスト(%s)は下記のネットワークのDNSサーバ(ネームサーバ)として運用されていなければなりません．</dt>\n", fqdn);
		ForLink(l, l_ns_network) {
			EdbTupleInfo	network = l->d;
			char	*cidr;
			int	status;
			struct	DNS_DOMAIN_INFO	dominfo;
			in_addr_t	address;
			int		mask;
			char	*net_fqdn = NULL;
			EdbTupleColumn	tc;
			EdbCatalogue	ca_ns;
			int	primary = FALSE;
			char	*master_ip;
			char	*revname = NULL;

			tc = edb_tuple_seek_column_by_xmlname(ec, network, (utf8_t *)"@.ns");
			if(tc==NULL) continue;
			ca_ns = edb_dns_lookup_iphosts_by_EID(ec, DatumEID(TCDatumFirst(tc)), ec->max_mapto_recurse);
			if(edb_catalogue_lookup(ca_ns, TupleEID(host))) primary = TRUE;
			if(edb_catalogue_size(ca_ns)>0) {
				master_ip = edb_dns_host_address(ec, (EdbTupleInfo)ca_ns->link->d);
			} else {
				master_ip = "XX.XX.XX.XX";
			}
			edb_catalogue_free(ca_ns);

			doc_puts(doc, "<dd>");
			cidr = edb_dns_network_cidr(ec, network);
			address = cidr_to_address(cidr);
			mask = cidr_to_mask(cidr);
			if(mask<=8) {
				asprintf(&net_fqdn, "%d.in-addr.arpa", (address>>24)&0xff);
				asprintf(&revname, "%d.rev", (address>>24)&0xff);
			} else if(mask<=16) {
				asprintf(&net_fqdn, "%d.%d.in-addr.arpa", (address>>16)&0xff, (address>>24)&0xff);
				asprintf(&revname, "%d.%d.rev", (address>>24)&0xff, (address>>16)&0xff);
			} else if(mask<=24) {
				asprintf(&net_fqdn, "%d.%d.%d.in-addr.arpa", (address>>8)&0xff, (address>>16)&0xff, (address>>24)&0xff);
				asprintf(&revname, "%d.%d.%d.rev", (address>>24)&0xff, (address>>16)&0xff, (address>>8)&0xff);
			} else {
				continue;
			}

			status = dns_query_domain_information(net_fqdn, fqdn, &dominfo, stderr);
			if(status>0) {
				doc_puts(doc, "◎");
			} else if(status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, "<span style=\"color:#800000\">△</span>");
			} else {
				doc_puts(doc, "<span style=\"color:red\">×</span>");
			}

			doc_puts(doc, " ");
			edb_dns_link_to_network(ec, network, 0, doc);
			doc_printf(doc, " (%s)", net_fqdn);
			doc_puts(doc, primary ? " (PRIMARY)" : " (secondary)");
			if(status>0) {
			} else if(status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, " <span style=\"color:#800000\">DNSサーバ(ネームサーバ)の設定がされていないようです．</span>");
			} else {
				doc_puts(doc, " <span style=\"color:red\">DNSサーバ(ネームサーバ)が応答しません．</span>");
			}
			doc_puts(doc, "</dd>\n");
			if(!primary && status==DNS_NOT_AUTHORITATIVE) {
				doc_puts(doc, "<dd><dl><dd><table border=\"1\"><tr><td>\n");
				doc_puts(doc, "このゾーンに関するnamed.conf (BIND)のセカンダリ設定は，<pre>");
				doc_printf(doc, "zone \"%s\" {\n", net_fqdn);
				doc_puts(doc, "	type slave;\n");
				doc_printf(doc, "	file \"back/%s\";\n", revname);
				doc_printf(doc, "	masters { %s; };\n", master_ip);
				doc_puts(doc, "};\n");
				doc_puts(doc, "</pre>のような記述になります．\n");
				doc_puts(doc, "</td></tr></table></dd></dl></dd>\n");
			}
			FREE(revname);
			FREE(net_fqdn);
			FREE(cidr);
			FREE(master_ip);
			doc_flush(doc);
		}
		link_free(l_ns_network);
	}

	l_mx_domain = NULL;
	ForLink(l, l_domain) {
		EdbTupleInfo	domain = l->d;
		EdbTupleColumn	tc;
		EdbCatalogue	ca_ns;

		tc = edb_tuple_seek_column_by_xmlname(ec, domain, (utf8_t *)"@.mx");
		if(tc==NULL) continue;
		ca_ns = edb_dns_lookup_iphosts_by_TC(ec, tc);
		if(edb_catalogue_lookup(ca_ns, TupleEID(host))) LinkAppend(l_mx_domain, link_atom_new(domain, NULL));
		edb_catalogue_free(ca_ns);
	}
	if(l_mx_domain) {
		doc_printf(doc, "<dt>◇ このホスト(%s)は下記のドメイン宛ての電子メールを受信し，正しく配送しなくてはなりません．</dt>\n", fqdn);
		ForLink(l, l_mx_domain) {
			EdbTupleInfo	domain = l->d;
			char	*dom_fqdn;
			int	status;
			struct	DNS_DOMAIN_INFO	dominfo;

			doc_puts(doc, "<dd>");
			dom_fqdn = edb_dns_domain_fqdn(ec, domain);
			doc_puts(doc, " ");
			edb_dns_link_to_domain(ec, domain, 0, doc);
			str_minuscule1(dom_fqdn);
			doc_printf(doc, "&nbsp;&nbsp;(宛先: <var style=\"color:green\">username</var>@<u>%s</u> )", dom_fqdn);
			doc_puts(doc, "</dd>\n");
			FREE(dom_fqdn);
			doc_flush(doc);
		}
		link_free(l_mx_domain);
	}

	doc_puts(doc, "</dl>");
	link_free(l_domain);
	link_free(l_network);

	doc_puts(doc, "</blockquote>\n");
	html_FOOT(ec, doc);
	return TRUE;
}

static
int
edb_dns_pki_host_certificate(EdbContext ec, char *fqdn, int text, FILE *fp)
{
	EdbTupleInfo	host;
	EdbColumnInfo	ci;
	EdbTupleColumn	tc;
	EdbDatum	datum;
	char	*cert_type, *s, *pem;
	Doc	doc;

	cert_type = strchr(fqdn, '/');
	if(cert_type) *cert_type++ = '\0';
	host = edb_dns_lookup_host_by_fqdn(ec, fqdn);
	if(host==NULL) {
		fprintf(fp, "; can't get information of host (%s).\n", fqdn);
		return FALSE;
	}

	ci = edb_pki_tuple_certifiable(ec, host);
	if(ci==NULL) return FALSE;
	tc = edb_tuple_seek_column(ec, host, ColumnXN(ci));
	if(tc==NULL || (datum=TCDatumFirst(tc))==NULL || !DatumEnglishIsValid(datum)) {
		fprintf(fp, "; host (%s) has no certificate.\n", fqdn);
		return FALSE;
	}
	if(!edb_pki_verify_certificate(ec, DatumEnglish(datum))) {
		fprintf(fp, "; host (%s) certificate was revoked.\n", fqdn);
		return FALSE;
	}
	doc = doc_fpopen(DocML_PLAIN, stdout);
	if(text) edb_pki_print_certificate_ex(ec, DatumEnglish(datum), doc);
	edb_pki_print_certificate(ec, DatumEnglish(datum), doc);
	doc_fpclose(doc);

	return TRUE;
}

static
int
edb_dns_pki_root_certificate(EdbContext ec, int text, FILE *stdout)
{
	Doc	doc;
	Link	l, l_str;
	utf8_t	*certificate;

	doc = doc_fpopen(DocML_PLAIN, stdout);
	l_str = file_load_as_linked_str(EDB_PKI_ROOT_CRT, NULL);
	if(l_str) {
		certificate = (utf8_t *)str_concatenate(l_str, ' ');
		if(certificate) {
			if(text) edb_pki_print_certificate_ex(ec, certificate, doc);
			edb_pki_print_certificate(ec, certificate, doc);
		}
		FREE(certificate);
	}
	link_free(l_str);
	doc_fpclose(doc);
	return TRUE;
}

static
int
edb_dns_pki_root_certificate2nd(EdbContext ec, int text, FILE *stdout)
{
	Doc	doc;
	Link	l, l_str;
	utf8_t	*certificate;

	doc = doc_fpopen(DocML_PLAIN, stdout);
	l_str = file_load_as_linked_str(EDB_PKI2_ROOT_CRT, NULL);
	if(l_str) {
		certificate = (utf8_t *)str_concatenate(l_str, ' ');
		if(certificate) {
			if(text) edb_pki_print_certificate_ex(ec, certificate, doc);
			edb_pki_print_certificate(ec, certificate, doc);
		}
		FREE(certificate);
	}
	link_free(l_str);
	doc_fpclose(doc);
	return TRUE;
}

/*#main--- main function */
int
main(int ac, char *av[])
{
	int i;
	int n;
	int sec;
	char *sectype;
	Link l, hosts;
	char	*query;
	char	*dbname = NULL;
	EdbContext	ec;

	prog_name = av[0];
	for(i=1; i<ac; ++i) {
		if(!strcmp(av[i], "-D") && i+1<ac) {
			dbname = av[i+1];
			++i;
		}
	}

	ec = edb_configure(UNIVERSITY_DBCONF);
	if(!ec->university) {
#if 0
		ec = edb_configure(PUBLIC_DBCONF);
#else
		printf("Content-Type: text/plain; charset=utf-8%c%c", 10, 10);
		printf("access denied\n");
		fprintf(stderr, "edb_dns_make: access denied\n");
		return FALSE;
#endif
	}
	if(dbname) edb_dbpath_set(ec, dbname, TRUE);

	if(!edb_open(ec)) {
		fprintf(stderr, "edb_dns_make: can't open database\n");
		return FALSE;
	}
	edb_trace_set(ec, stderr);

	edb_begin(ec, FALSE);

	query = getenv("PATH_INFO");
	if(query==NULL) {
		printf("Content-Type: text/plain; charset=utf-8%c%c", 10, 10);
		printf("usage: /DNS/BIND/ZONE/domain\n");
		printf("usage: /DNS/BIND/REV/netaddress\n");
		exit(1);
	}
	query = STRDUP(query);
	if(!strncmp(query, "/BIND/ZONE/", 11)) {
		char *domainname;

		printf("Content-Type: text/plain; charset=utf-8%c%c", 10, 10);
		if(ec->university) {
			domainname = strdup(&query[11]);
			str_capitalize1(domainname);
			edb_dns_make_zoneinfo(ec, domainname, stdout);
			FREE(domainname);
		} else {
			printf("Access denied.\n");
		}
	} else if(!strncmp(query, "/BIND/REV/", 10)) {
		char	*network;

		printf("Content-Type: text/plain; charset=utf-8%c%c", 10, 10);
		if(ec->university) {
			if(strchr(&query[10], '/')) {	/* CIDR form */
				asprintf(&network, "%s", &query[10]);
			} else {
				asprintf(&network, "%s/24", &query[10]);
			}
			str_capitalize1(network);
			edb_dns_make_revinfo(ec, network, stdout);
			FREE(network);
		} else {
			printf("Access denied.\n");
		}
	} else if(!strncmp(query, "/BIND/HOST/", 11)) {
		char	*fqdn;

		printf("Content-Type: text/html; charset=utf-8%c%c", 10, 10);
		if(ec->university) {
			Doc	doc;

			asprintf(&fqdn, "%s", &query[11]);
			doc = doc_fpopen(DocML_HTML, stdout);
			str_minuscule1(fqdn);
			edb_dns_host_info(ec, fqdn, doc);
			doc_close(doc);
			FREE(fqdn);
		} else {
			printf("Access denied.\n");
		}
	} else if(!strncmp(query, "/BIND/CONF/", 11)) {
		char	*nameserver;

		printf("Content-Type: text/plain; charset=utf-8%c%c", 10, 10);
		if(ec->university) {
			nameserver = strdup(&query[11]);
			str_minuscule1(nameserver);
			edb_dns_make_nsconf(ec, nameserver, stdout);
			FREE(nameserver);
		} else {
			printf("Access denied.\n");
		}
	} else if(!strncmp(query, "/BIND/SCRIPT/", 13)) {
		char	*scriptname;

		printf("Content-Type: text/plain; charset=utf-8%c%c", 10, 10);
		scriptname = strdup(&query[13]);
		edb_dns_script_print(ec, scriptname, stdout);
		FREE(scriptname);
	} else if(!strncmp(query, "/LOOK/", 6)) {
		char	*look;
		Link	l, lstr;
		int	i, look_ac;
		char	**look_av;

		printf("Content-Type: text/plain; charset=utf-8%c%c", 10, 10);
		look = &query[6];
		lstr = str_split(look, '/');
		look_ac = link_atoms(lstr);
		look_av = (char **)malloc(sizeof(char *)*look_ac);
		for(i=0, l=lstr; l; ++i, l=l->n) look_av[i] = l->d;
		dns_look(ec, look_ac, look_av);
		link_free(lstr);
		free(look_av);
	} else if(!strcmp(query, "/DOMAIN/SUMMARY")) {
		Doc	doc;

		printf("Content-Type: text/html; charset=utf-8%c%c", 10, 10);
		doc = doc_fpopen(DocML_HTML, stdout);
		edb_dns_domain_summary(ec, doc);
		doc_close(doc);
	} else if(!strcmp(query, "/NETWORK/SUMMARY")) {
		Doc	doc;

		printf("Content-Type: text/html; charset=utf-8%c%c", 10, 10);
		doc = doc_fpopen(DocML_HTML, stdout);
		edb_dns_network_summary(ec, doc);
		doc_close(doc);
	} else if(!strncmp(query, "/DOMAIN/", 8)) {
		char	*fqdn;

		printf("Content-Type: text/html; charset=utf-8%c%c", 10, 10);
		if(ec->university) {
			Doc	doc;

			asprintf(&fqdn, "%s", &query[8]);
			doc = doc_fpopen(DocML_HTML, stdout);
			str_capitalize1(fqdn);
			edb_dns_domain_info(ec, fqdn, doc);
			doc_close(doc);
			FREE(fqdn);
		} else {
			printf("Access denied.\n");
		}
	} else if(!strncmp(query, "/NETWORK/", 9)) {
		char	*cidr;

		printf("Content-Type: text/html; charset=utf-8%c%c", 10, 10);
		if(ec->university) {
			Doc	doc;

			asprintf(&cidr, "%s", &query[9]);
			doc = doc_fpopen(DocML_HTML, stdout);
			edb_dns_network_info(ec, cidr, doc);
			doc_close(doc);
			FREE(cidr);
		} else {
			printf("Access denied.\n");
		}
	} else if(!strncmp(query, "/HOST/", 6)) {
		char	*fqdn;

		printf("Content-Type: text/html; charset=utf-8%c%c", 10, 10);
		if(ec->university) {
			Doc	doc;

			asprintf(&fqdn, "%s", &query[6]);
			doc = doc_fpopen(DocML_HTML, stdout);
			str_minuscule1(fqdn);
			edb_dns_host_info(ec, fqdn, doc);
			doc_close(doc);
			FREE(fqdn);
		} else {
			printf("Access denied.\n");
		}
	} else if(!strncmp(query, "/PKI/HOST/", 10) && strstr(query, "/certificate.")) {
		char	*host;

		printf("Content-Type: application/octet-stream%c%c", 10, 10);
		host = strdup(&query[10]);
		str_minuscule1(host);
		edb_dns_pki_host_certificate(ec, host, FALSE, stdout);
		FREE(host);
	} else if(!strcmp(query, "/PKI/CA/root.pem") || !strcmp(query, "/PKI/CA/root.crt")) {
		printf("Content-Type: application/octet-stream%c%c", 10, 10);
		edb_dns_pki_root_certificate(ec, FALSE, stdout);
	} else if(!strcmp(query, "/PKI/CA/root2nd.pem") || !strcmp(query, "/PKI/CA/root2nd.crt")) {
		printf("Content-Type: application/octet-stream%c%c", 10, 10);
		edb_dns_pki_root_certificate2nd(ec, FALSE, stdout);
	} else {
		printf("Content-Type: text/plain; charset=utf-8%c%c", 10, 10);
		printf("usage: /DNS/BIND/ZONE/domain\n");
		printf("usage: /DNS/BIND/REV/netaddress\n");
		printf("usage: /DNS/BIND/CONF/fqdn.of.name.server\n");
		printf("usage: /DNS/BIND/SCRIPT/name-of-script\n");
		printf("usage: /DNS/LOOK/query\n");
		printf("usage: /DNS/DOMAIN/SUMMARY\n");
		printf("usage: /DNS/NETWORK/SUMMARY\n");
		printf("usage: /DNS/HOST/fqdn.of.host.name\n");
		printf("usage: /DNS/PKI/HOST/fqdn.of.server/certificate.pem or .crt\n");
		printf("usage: /DNS/PKI/CA/root.pem or .crt\n");
		printf("usage: /DNS/PKI/CA/root2nd.pem or .crt\n");
		exit(1);
	}
	edb_end(ec);
	edb_close(ec);
	fflush(stdout);
	exit(0);
}


